from __future__ import print_function

import sys
from os import environ
from base64 import b64encode
import json
from datetime import datetime, timedelta
from itertools import chain

import requests
from urlobject import URLObject
import six
from six.moves.urllib.parse import urlencode
from nylas._client_sdk_version import __VERSION__
from nylas.client.delta_collection import DeltaCollection
from nylas.client.errors import MessageRejectedError, NylasApiError, RateLimitError
from nylas.client.outbox_models import Outbox
from nylas.client.restful_model_collection import RestfulModelCollection
from nylas.client.restful_models import (
    Calendar,
    Contact,
    Event,
    RoomResource,
    Message,
    Thread,
    File,
    Account,
    APIAccount,
    SingletonAccount,
    Folder,
    Label,
    Draft,
    Component,
    JobStatus,
    Webhook,
)
from nylas.client.neural_api_models import Neural
from nylas.client.scheduler_restful_model_collection import (
    SchedulerRestfulModelCollection,
)
from nylas.client.authentication_models import Authentication
from nylas.utils import timestamp_from_dt, create_request_body, AuthMethod, HttpMethod

DEBUG = environ.get("NYLAS_CLIENT_DEBUG")
API_SERVER = "https://api.nylas.com"
SUPPORTED_API_VERSION = "2.5"


def _validate(response):
    if DEBUG:  # pragma: no cover
        print(
            "{method} {url} ({body}) => {status}: {text}".format(
                method=response.request.method,
                url=response.request.url,
                body=response.request.body,
                status=response.status_code,
                text=response.text,
            )
        )

    if response.status_code == 402:
        # HTTP status code 402 normally means "Payment Required",
        # but when Nylas uses that status code, it means something different.
        # Usually it indicates an upstream error on the provider.
        # We let Requests handle most HTTP errors, but for this one,
        # we will handle it separate and handle a _different_ exception
        # so that users don't think they need to pay.
        raise MessageRejectedError(response)
    elif response.status_code == 429:
        raise RateLimitError(response)
    elif response.status_code >= 400:
        raise NylasApiError(response)

    return response


def _validate_availability_query(query):
    if (query.get("emails", None) is None or len(query["emails"]) == 0) and (
        query.get("calendars", None) is None or len(query["calendars"]) == 0
    ):
        raise ValueError("Must set either 'emails' or 'calendars' in the query.")


class APIClient(json.JSONEncoder):
    """API client for the Nylas API."""

    def __init__(
        self,
        client_id=environ.get("NYLAS_CLIENT_ID"),
        client_secret=environ.get("NYLAS_CLIENT_SECRET"),
        access_token=environ.get("NYLAS_ACCESS_TOKEN"),
        api_server=API_SERVER,
        api_version=SUPPORTED_API_VERSION,
    ):
        if not api_server.startswith("https://"):
            raise Exception(
                "When overriding the Nylas API server address, you"
                " must include https://"
            )
        self.api_server = api_server
        self.api_version = api_version
        self.authorize_url = api_server + "/oauth/authorize"
        self.access_token_url = api_server + "/oauth/token"
        self.revoke_url = api_server + "/oauth/revoke"
        self.application_url = api_server + "/a/{client_id}"
        self.revoke_all_url = self.application_url + "/accounts/{account_id}/revoke-all"
        self.ip_addresses_url = api_server + "/a/{client_id}/ip_addresses"
        self.token_info_url = self.application_url + "/accounts/{account_id}/token-info"

        self.client_secret = client_secret
        self.client_id = client_id

        self.session = requests.Session()
        self.version = __VERSION__
        major, minor, revision, _, __ = sys.version_info
        version_header = "Nylas Python SDK {} - {}.{}.{}".format(
            self.version, major, minor, revision
        )
        self.session.headers = {
            "X-Nylas-API-Wrapper": "python",
            "X-Nylas-Client-Id": self.client_id,
            "Nylas-API-Version": self.api_version,
            "User-Agent": version_header,
        }
        self._access_token = None
        self.access_token = access_token
        self.auth_token = None

        # Requests to the /a/ namespace don't use an auth token but
        # the client_secret. Set up a specific session for this.
        self.admin_session = requests.Session()

        if client_secret is not None:
            self.admin_session.headers = {
                "X-Nylas-API-Wrapper": "python",
                "X-Nylas-Client-Id": self.client_id,
                "Nylas-API-Version": self.api_version,
                "User-Agent": version_header,
            }
            self.admin_session.headers.update(self._add_auth_header(AuthMethod.BASIC))
        super(APIClient, self).__init__()

    @property
    def access_token(self):
        return self._access_token

    @access_token.setter
    def access_token(self, value):
        self._access_token = value

    def authentication_url(
        self,
        redirect_uri,
        login_hint="",
        state="",
        scopes=("email", "calendar", "contacts"),
        provider="",
        redirect_on_error=None,
    ):
        args = {
            "redirect_uri": redirect_uri,
            "client_id": self.client_id or "None",  # 'None' for back-compat
            "response_type": "code",
            "login_hint": login_hint,
            "state": state,
        }

        if scopes:
            if isinstance(scopes, str):
                scopes = [scopes]
            args["scopes"] = ",".join(scopes)
        if provider and provider in [
            "icloud",
            "gmail",
            "office365",
            "exchange",
            "imap",
        ]:
            args["provider"] = provider
        if redirect_on_error is not None and isinstance(redirect_on_error, bool):
            args["redirect_on_error"] = "true" if redirect_on_error is True else "false"

        url = URLObject(self.authorize_url).add_query_params(args.items())
        return str(url)

    def send_authorization(self, code):
        """
        Exchanges an authorization code for an access token.

        Args:
            code (str): The authorization code returned from authenticating the user

        Returns:
            dict: The response from the API containing the access token
        """
        args = {
            "client_id": self.client_id,
            "client_secret": self.client_secret,
            "grant_type": "authorization_code",
            "code": code,
        }

        headers = {
            "Content-type": "application/x-www-form-urlencoded",
            "Accept": "application/json",
        }

        resp = self._request(
            HttpMethod.POST,
            self.access_token_url,
            headers=headers,
            data=urlencode(args),
        )
        results = _validate(resp).json()

        self.access_token = results["access_token"]
        return resp

    def token_for_code(self, code):
        self.send_authorization(code)
        return self.access_token

    def is_opensource_api(self):
        if self.client_id is None and self.client_secret is None:
            return True

        return False

    def application_details(self):
        application_details_url = self.application_url.format(client_id=self.client_id)
        resp = self.admin_session.get(application_details_url)
        _validate(resp).json()
        return resp.json()

    def update_application_details(
        self, application_name=None, icon_url=None, redirect_uris=None
    ):
        application_details_url = self.application_url.format(client_id=self.client_id)
        data = {}
        if application_name is not None:
            data["application_name"] = application_name
        if icon_url is not None:
            data["icon_url"] = icon_url
        if redirect_uris is not None:
            data["redirect_uris"] = redirect_uris

        headers = {"Content-Type": "application/json"}
        headers.update(self.admin_session.headers)
        resp = self.admin_session.put(
            application_details_url, json=data, headers=headers
        )
        return _validate(resp).json()

    def revoke_token(self):
        resp = requests.post(self.revoke_url, auth=(self.access_token, None))
        _validate(resp)
        self.auth_token = None
        self.access_token = None

    def revoke_all_tokens(self, keep_access_token=None):
        revoke_all_url = self.revoke_all_url.format(
            client_id=self.client_id, account_id=self.account.id
        )
        data = {}
        if keep_access_token is not None:
            data["keep_access_token"] = keep_access_token

        headers = {"Content-Type": "application/json"}
        headers.update(self.admin_session.headers)
        resp = self.admin_session.post(revoke_all_url, json=data, headers=headers)
        _validate(resp).json()
        if keep_access_token != self.access_token:
            self.auth_token = None
            self.access_token = None

    def ip_addresses(self):
        ip_addresses_url = self.ip_addresses_url.format(client_id=self.client_id)
        resp = self.admin_session.get(ip_addresses_url)
        _validate(resp).json()
        return resp.json()

    def token_info(self):
        token_info_url = self.token_info_url.format(
            client_id=self.client_id, account_id=self.account.id
        )
        headers = {"Content-Type": "application/json"}
        headers.update(self.admin_session.headers)
        resp = self.admin_session.post(
            token_info_url, headers=headers, json={"access_token": self.access_token}
        )
        _validate(resp).json()
        return resp.json()

    def free_busy(self, emails, start_at, end_at, calendars=None):
        if isinstance(emails, six.string_types):
            emails = [emails]
        if isinstance(start_at, datetime):
            start_time = timestamp_from_dt(start_at)
        else:
            start_time = start_at
        if isinstance(end_at, datetime):
            end_time = timestamp_from_dt(end_at)
        else:
            end_time = end_at
        url = "{api_server}/calendars/free-busy".format(api_server=self.api_server)
        data = {
            "emails": emails,
            "start_time": start_time,
            "end_time": end_time,
        }
        if calendars is not None and len(calendars) > 0:
            data["calendars"] = calendars

        _validate_availability_query(data)
        resp = self._request(HttpMethod.POST, url, json=data, cls=Calendar)
        _validate(resp)
        return resp.json()

    def open_hours(self, emails, days, timezone, start, end):
        if isinstance(emails, six.string_types):
            emails = [emails]
        if isinstance(days, int):
            days = [days]
        if isinstance(start, datetime):
            start = "{hour}:{minute}".format(hour=start.hour, minute=start.minute)
        if isinstance(start, datetime):
            end = "{hour}:{minute}".format(hour=end.hour, minute=end.minute)
        return {
            "emails": emails,
            "days": days,
            "timezone": timezone,
            "start": start,
            "end": end,
            "object_type": "open_hours",
        }

    def availability(
        self,
        emails,
        duration,
        interval,
        start_at,
        end_at,
        event_collection_id=None,
        buffer=None,
        round_robin=None,
        free_busy=None,
        open_hours=None,
        calendars=None,
    ):
        if isinstance(emails, six.string_types):
            emails = [emails]
        if isinstance(duration, timedelta):
            duration_minutes = int(duration.total_seconds() // 60)
        else:
            duration_minutes = int(duration)
        if isinstance(interval, timedelta):
            interval_minutes = int(interval.total_seconds() // 60)
        else:
            interval_minutes = int(interval)
        if isinstance(start_at, datetime):
            start_time = timestamp_from_dt(start_at)
        else:
            start_time = start_at
        if isinstance(end_at, datetime):
            end_time = timestamp_from_dt(end_at)
        else:
            end_time = end_at
        if open_hours is not None:
            self._validate_open_hours(emails, open_hours, free_busy)

        url = "{api_server}/calendars/availability".format(api_server=self.api_server)
        data = {
            "emails": emails,
            "duration_minutes": duration_minutes,
            "interval_minutes": interval_minutes,
            "start_time": start_time,
            "end_time": end_time,
            "free_busy": free_busy or [],
            "open_hours": open_hours or [],
        }
        if buffer is not None:
            data["buffer"] = buffer
        if round_robin is not None:
            data["round_robin"] = round_robin
        if event_collection_id is not None:
            data["event_collection_id"] = event_collection_id
        if calendars is not None and len(calendars) > 0:
            data["calendars"] = calendars

        _validate_availability_query(data)
        resp = self._request(HttpMethod.POST, url, json=data, cls=Calendar)
        _validate(resp)
        return resp.json()

    def consecutive_availability(
        self,
        emails,
        duration,
        interval,
        start_at,
        end_at,
        buffer=None,
        free_busy=None,
        open_hours=None,
        calendars=None,
    ):
        if isinstance(emails, six.string_types):
            emails = [[emails]]
        elif len(emails) > 0 and isinstance(emails[0], list) is False:
            raise ValueError("'emails' must be a list of lists.")
        if isinstance(duration, timedelta):
            duration_minutes = int(duration.total_seconds() // 60)
        else:
            duration_minutes = int(duration)
        if isinstance(interval, timedelta):
            interval_minutes = int(interval.total_seconds() // 60)
        else:
            interval_minutes = int(interval)
        if isinstance(start_at, datetime):
            start_time = timestamp_from_dt(start_at)
        else:
            start_time = start_at
        if isinstance(end_at, datetime):
            end_time = timestamp_from_dt(end_at)
        else:
            end_time = end_at
        if open_hours is not None:
            self._validate_open_hours(emails, open_hours, free_busy)

        url = "{api_server}/calendars/availability/consecutive".format(
            api_server=self.api_server
        )
        data = {
            "emails": emails,
            "duration_minutes": duration_minutes,
            "interval_minutes": interval_minutes,
            "start_time": start_time,
            "end_time": end_time,
            "free_busy": free_busy or [],
            "open_hours": open_hours or [],
        }
        if buffer is not None:
            data["buffer"] = buffer
        if calendars is not None and len(calendars) > 0:
            data["calendars"] = calendars

        _validate_availability_query(data)
        resp = self._request(HttpMethod.POST, url, json=data, cls=Calendar)
        _validate(resp)
        return resp.json()

    @property
    def account(self):
        return self._get_resource(SingletonAccount, "")

    @property
    def accounts(self):
        if self.is_opensource_api():
            return RestfulModelCollection(APIAccount, self)
        return RestfulModelCollection(Account, self)

    @property
    def threads(self):
        return RestfulModelCollection(Thread, self)

    @property
    def folders(self):
        return RestfulModelCollection(Folder, self)

    @property
    def labels(self):
        return RestfulModelCollection(Label, self)

    @property
    def messages(self):
        return RestfulModelCollection(Message, self)

    @property
    def files(self):
        return RestfulModelCollection(File, self)

    @property
    def drafts(self):
        return RestfulModelCollection(Draft, self)

    @property
    def contacts(self):
        return RestfulModelCollection(Contact, self)

    @property
    def events(self):
        return RestfulModelCollection(Event, self)

    @property
    def room_resources(self):
        return RestfulModelCollection(RoomResource, self)

    @property
    def calendars(self):
        return RestfulModelCollection(Calendar, self)

    @property
    def job_statuses(self):
        return RestfulModelCollection(JobStatus, self)

    @property
    def scheduler(self):
        return SchedulerRestfulModelCollection(self)

    @property
    def components(self):
        return RestfulModelCollection(Component, self)

    @property
    def deltas(self):
        return DeltaCollection(self)

    @property
    def webhooks(self):
        return RestfulModelCollection(Webhook, self)

    @property
    def neural(self):
        return Neural(self)

    @property
    def outbox(self):
        return Outbox(self)

    @property
    def authentication(self):
        return Authentication(self)

    ##########################################################
    #   Private functions used by Restful Model Collection   #
    ##########################################################

    def _get_http_session(self, api_root):
        # Is this a request for a resource under the accounts/billing/admin
        # namespace (/a)? If the latter, pass the client_secret
        # instead of the secret_token
        if api_root:
            return self.admin_session
        return self.session

    def _get_resources(self, cls, extra=None, **filters):
        # FIXME @karim: remove this interim code when we've got rid
        # of the old accounts API.
        postfix = "/{}".format(extra) if extra else ""
        path = "/{}".format(cls.collection_name) if cls.collection_name else ""
        if not cls.api_root:
            url = "{server}{path}{postfix}".format(
                server=self.api_server, path=path, postfix=postfix
            )
        else:
            url = "{server}/{prefix}/{client_id}{path}{postfix}".format(
                server=self.api_server,
                prefix=cls.api_root,
                client_id=self.client_id,
                path=path,
                postfix=postfix,
            )

        converted_data = create_request_body(filters, cls.datetime_filter_attrs)
        url = str(URLObject(url).add_query_params(converted_data.items()))
        response = self._request(HttpMethod.GET, url, cls=cls)
        results = _validate(response).json()
        return [cls.create(self, **x) for x in results if x is not None]

    def _get_resource_raw(
        self,
        cls,
        resource_id,
        extra=None,
        headers=None,
        stream=False,
        path=None,
        stream_timeout=None,
        **filters
    ):
        """Get an individual REST resource"""
        if path is None:
            path = cls.collection_name
        postfix = "/{}".format(extra) if extra else ""
        path = "/{}".format(path) if path else ""
        resource_id = "/{}".format(resource_id) if resource_id else ""
        if not cls.api_root:
            url = "{server}{path}{id}{postfix}".format(
                server=self.api_server, path=path, id=resource_id, postfix=postfix
            )
        else:
            url = "{server}/{prefix}/{client_id}{path}{id}{postfix}".format(
                server=self.api_server,
                prefix=cls.api_root,
                client_id=self.client_id,
                path=path,
                id=resource_id,
                postfix=postfix,
            )

        converted_data = create_request_body(filters, cls.datetime_filter_attrs)
        url = str(URLObject(url).add_query_params(converted_data.items()))

        response = self._request(
            HttpMethod.GET,
            url,
            cls=cls,
            headers=headers,
            stream=stream,
            timeout=stream_timeout,
        )
        return _validate(response)

    def _get_resource(self, cls, resource_id, **filters):
        response = self._get_resource_raw(cls, resource_id, **filters)
        result = response.json()
        if isinstance(result, list):
            result = result[0]
        return cls.create(self, **result)

    def _get_resource_data(self, cls, resource_id, extra=None, headers=None, **filters):
        response = self._get_resource_raw(
            cls, resource_id, extra=extra, headers=headers, **filters
        )
        return response.content

    def _create_resource(self, cls, data, **kwargs):
        name = "{prefix}{path}".format(
            prefix="/{}/{}".format(cls.api_root, self.client_id)
            if cls.api_root
            else "",
            path="/{}".format(cls.collection_name) if cls.collection_name else "",
        )
        url = (
            URLObject(self.api_server)
            .with_path("{name}".format(name=name))
            .set_query_params(**kwargs)
        )

        if cls == File:
            response = self._request(HttpMethod.POST, url, cls=cls, files=data)
        else:
            converted_data = create_request_body(data, cls.datetime_attrs)
            headers = {"Content-Type": "application/json"}
            response = self._request(
                HttpMethod.POST, url, cls=cls, headers=headers, json=converted_data
            )

        result = _validate(response).json()
        if cls.collection_name == "send":
            return result
        return cls.create(self, **result)

    def _create_resources(self, cls, data):
        name = "{prefix}{path}".format(
            prefix="/{}/{}".format(cls.api_root, self.client_id)
            if cls.api_root
            else "",
            path="/{}".format(cls.collection_name) if cls.collection_name else "",
        )
        url = URLObject(self.api_server).with_path("{name}".format(name=name))

        if cls == File:
            response = self._request(HttpMethod.POST, url, cls=cls, files=data)
        else:
            converted_data = [
                create_request_body(datum, cls.datetime_attrs) for datum in data
            ]
            headers = {"Content-Type": "application/json"}
            response = self._request(
                HttpMethod.POST, url, cls=cls, headers=headers, json=converted_data
            )

        results = _validate(response).json()
        return [cls.create(self, **x) for x in results]

    def _delete_resource(self, cls, resource_id, data=None, **kwargs):
        name = "{prefix}{path}".format(
            prefix="/{}/{}".format(cls.api_root, self.client_id)
            if cls.api_root
            else "",
            path="/{}".format(cls.collection_name) if cls.collection_name else "",
        )
        url = (
            URLObject(self.api_server)
            .with_path("{name}/{id}".format(name=name, id=resource_id))
            .set_query_params(**kwargs)
        )
        if data:
            _validate(self._request(HttpMethod.DELETE, url, cls=cls, json=data))
        else:
            _validate(self._request(HttpMethod.DELETE, url, cls=cls))

    def _request_update_resource(
        self, method, cls, resource_id, data, extra=None, path=None, **kwargs
    ):
        if path is None:
            path = cls.collection_name
        name = "{prefix}{path}".format(
            prefix="/{}/{}".format(cls.api_root, self.client_id)
            if cls.api_root
            else "",
            path="/{}".format(path) if path else "",
        )

        postfix = "/{}".format(extra) if extra else ""
        url = (
            URLObject(self.api_server)
            .with_path(
                "{name}/{id}{postfix}".format(
                    name=name, id=resource_id, postfix=postfix
                )
            )
            .set_query_params(**kwargs)
        )
        converted_data = create_request_body(data, cls.datetime_attrs)

        response = self._request(method, url, cls=cls, json=converted_data)

        result = _validate(response)
        return result.json()

    def _patch_resource(self, cls, resource_id, data, extra=None, path=None, **kwargs):
        return self._request_update_resource(
            HttpMethod.PATCH, cls, resource_id, data, extra=extra, path=path, **kwargs
        )

    def _put_resource(self, cls, resource_id, data, extra=None, path=None, **kwargs):
        return self._request_update_resource(
            HttpMethod.PUT, cls, resource_id, data, extra=extra, path=path, **kwargs
        )

    def _update_resource(self, cls, resource_id, data, **kwargs):
        result = self._put_resource(cls, resource_id, data, **kwargs)
        return cls.create(self, **result)

    def _post_resource(self, cls, resource_id, method_name, data, path=None):
        if path is None:
            path = cls.collection_name
        path = "/{}".format(path) if path else ""
        resource_id = "/{}".format(resource_id) if resource_id else ""
        method = "/{}".format(method_name) if method_name else ""
        if not cls.api_root:
            url_path = "{name}{id}{method}".format(
                name=path, id=resource_id, method=method
            )
        else:
            # Management method.
            url_path = "/{prefix}/{client_id}{path}{id}{method}".format(
                prefix=cls.api_root,
                client_id=self.client_id,
                path=path,
                id=resource_id,
                method=method,
            )

        url = URLObject(self.api_server).with_path(url_path)
        converted_data = create_request_body(data, cls.datetime_attrs)

        response = self._request(HttpMethod.POST, url, cls=cls, json=converted_data)
        return _validate(response).json()

    def _call_resource_method(self, cls, resource_id, method_name, data):
        """POST a dictionary to an API method,
        for example /a/.../accounts/id/upgrade"""

        result = self._post_resource(cls, resource_id, method_name, data)
        return cls.create(self, **result)

    def _request_neural_resource(self, cls, data, path=None, method=None):
        if path is None:
            path = cls.collection_name
        if method is None:
            method = HttpMethod.PUT
        url = URLObject(self.api_server).with_path("/neural/{name}".format(name=path))

        converted_data = create_request_body(data, cls.datetime_attrs)
        response = self._request(method, url, cls=cls, json=converted_data)

        result = _validate(response).json()
        if isinstance(result, list):
            object_list = []
            for obj in result:
                object_list.append(cls.create(self, **obj))
            return object_list

        return cls.create(self, **result)

    def _validate_open_hours(self, emails, open_hours, free_busy):
        if isinstance(open_hours, list) is False:
            raise ValueError("'open_hours' must be an array.")
        open_hours_emails = list(
            chain.from_iterable([oh["emails"] for oh in open_hours])
        )
        free_busy_emails = (
            [fb["email"] for fb in free_busy] if free_busy is not None else []
        )
        if isinstance(emails[0], list) is True:
            emails = list(chain.from_iterable(emails))
        for email in open_hours_emails:
            if (email in emails) is False and (email in free_busy_emails) is False:
                raise ValueError(
                    "Open Hours cannot contain an email not present in the main email list or the free busy email list."
                )

    def _request(self, method, url, cls=None, headers=None, **kwargs):
        api_root = None
        auth_method = None
        if cls:
            api_root = cls.api_root
            auth_method = cls.auth_method

        session = self._get_http_session(api_root)
        headers = headers or {}
        headers.update(session.headers)
        headers.update(self._add_auth_header(auth_method))
        return session.request(method.name, url, headers=headers, **kwargs)

    def _add_auth_header(self, auth_method):
        authorization = None
        if auth_method == AuthMethod.BEARER:
            authorization = (
                "Bearer {token}".format(token=self.access_token)
                if self.access_token
                else None
            )
        elif auth_method == AuthMethod.BASIC_CLIENT_ID_AND_SECRET:
            if self.client_id and self.client_secret:
                credential = "{client_id}:{client_secret}".format(
                    client_id=self.client_id, client_secret=self.client_secret
                )
                authorization = "Basic {credential}".format(
                    credential=b64encode(credential.encode("utf8")).decode("utf8")
                )
        else:
            if self.client_secret:
                b64_client_secret = b64encode(
                    ("{}:".format(self.client_secret)).encode("utf8")
                )
                authorization = "Basic {secret}".format(
                    secret=b64_client_secret.decode("utf8")
                )

        return {"Authorization": authorization} if authorization else {}



import pytest
from nylas.client.restful_models import Folder, Thread, Message


@pytest.mark.usefixtures("mock_folder")
def test_get_change_folder(api_client):
    folder = api_client.folders.get("anuep8pe5ug3xrupchwzba2o8")
    assert folder is not None
    assert isinstance(folder, Folder)
    assert folder.display_name == "My Folder"
    folder.display_name = "My New Folder"
    folder.save()
    assert folder.display_name == "My New Folder"


@pytest.mark.usefixtures("mock_folder", "mock_threads")
def test_folder_threads(api_client):
    folder = api_client.folders.get("anuep8pe5ug3xrupchwzba2o8")
    assert folder.threads
    assert all(isinstance(thread, Thread) for thread in folder.threads)


@pytest.mark.usefixtures("mock_folder", "mock_messages")
def test_folder_messages(api_client):
    folder = api_client.folders.get("anuep8pe5ug3xrupchwzba2o8")
    assert folder.messages
    assert all(isinstance(message, Message) for message in folder.messages)



import json
from datetime import datetime

import pytest
from nylas.client.restful_models import File

from nylas.client.neural_api_models import (
    NeuralSignatureContact,
    NeuralMessageOptions,
    Categorize,
    NeuralCategorizer,
)


@pytest.mark.usefixtures("mock_sentiment_analysis")
def test_sentiment_analysis_message(mocked_responses, api_client, account_id):
    analysis_response = api_client.neural.sentiment_analysis_message(["message_id"])
    assert len(mocked_responses.calls) == 1
    request = mocked_responses.calls[0].request
    assert json.loads(request.body) == {"message_id": ["message_id"]}
    assert len(analysis_response) == 1
    analysis = analysis_response[0]
    assert analysis.account_id == account_id
    assert analysis.processed_length == 11
    assert analysis.sentiment == "NEUTRAL"
    assert analysis.sentiment_score == 0.30000001192092896
    assert analysis.text == "hello world"


@pytest.mark.usefixtures("mock_sentiment_analysis")
def test_sentiment_analysis_text(mocked_responses, api_client, account_id):
    analysis = api_client.neural.sentiment_analysis_text("hello world")
    assert len(mocked_responses.calls) == 1
    request = mocked_responses.calls[0].request
    assert json.loads(request.body) == {"text": "hello world"}
    assert analysis.account_id == account_id
    assert analysis.processed_length == 11
    assert analysis.sentiment == "NEUTRAL"
    assert analysis.sentiment_score == 0.30000001192092896
    assert analysis.text == "hello world"


@pytest.mark.usefixtures("mock_extract_signature")
def test_extract_signature(mocked_responses, api_client):
    signature_response = api_client.neural.extract_signature(["abc123"])
    assert len(mocked_responses.calls) == 1
    request = mocked_responses.calls[0].request
    assert json.loads(request.body) == {"message_id": ["abc123"]}
    assert len(signature_response) == 1
    signature = signature_response[0]
    assert (
        signature.signature
        == "Nylas Swag\n\nSoftware Engineer\n\n123-456-8901\n\nswag@nylas.com"
    )
    assert signature.model_version == "0.0.1"
    assert isinstance(signature.contacts, NeuralSignatureContact)
    contact = signature.contacts
    assert contact.job_titles == ["Software Engineer"]
    assert contact.links == [
        {
            "description": "string",
            "url": "https://example.com/link.html",
        }
    ]
    assert contact.phone_numbers == ["123-456-8901"]
    assert contact.emails == ["swag@nylas.com"]
    assert contact.names == [
        {
            "first_name": "Nylas",
            "last_name": "Swag",
        }
    ]


@pytest.mark.usefixtures("mock_extract_signature")
def test_extract_signature_options(mocked_responses, api_client):
    options = NeuralMessageOptions(False, False, False, False, False)
    api_client.neural.extract_signature(["abc123"], False, options)
    request = mocked_responses.calls[0].request
    assert json.loads(request.body) == {
        "message_id": ["abc123"],
        "parse_contacts": False,
        "ignore_links": False,
        "ignore_images": False,
        "ignore_tables": False,
        "remove_conclusion_phrases": False,
        "images_as_markdowns": False,
    }


@pytest.mark.usefixtures("mock_extract_signature")
def test_signature_convert_contact(mocked_responses, api_client):
    signature = api_client.neural.extract_signature(["abc123"])
    contact = signature[0].contacts.to_contact_object()
    assert contact.given_name == "Nylas"
    assert contact.surname == "Swag"
    assert contact.job_title == "Software Engineer"
    assert len(contact.emails) == 1
    assert contact.emails["personal"] == ["swag@nylas.com"]
    assert len(contact.phone_numbers) == 1
    assert contact.phone_numbers["mobile"] == ["123-456-8901"]
    assert len(contact.web_pages) == 1
    assert contact.web_pages["string"] == ["https://example.com/link.html"]


@pytest.mark.usefixtures("mock_categorize")
def test_categorize(mocked_responses, api_client):
    response = api_client.neural.categorize(["abc123"])
    assert len(mocked_responses.calls) == 1
    request = mocked_responses.calls[0].request
    assert json.loads(request.body) == {"message_id": ["abc123"]}
    assert len(response) == 1
    assert isinstance(response[0].categorizer, Categorize)
    categorize = response[0].categorizer
    assert categorize.category == "feed"
    assert categorize.model_version == "6194f733"
    assert categorize.subcategories == ["ooo"]
    assert categorize.categorized_at == datetime.utcfromtimestamp(1627076720)


@pytest.mark.usefixtures("mock_categorize")
def test_recategorize(mocked_responses, api_client):
    categorize = api_client.neural.categorize("abc123")
    recategorize = categorize[0].recategorize("conversation")
    assert len(mocked_responses.calls) == 3
    request = mocked_responses.calls[1].request
    assert json.loads(request.body) == {
        "message_id": "abc123",
        "category": "conversation",
    }
    assert isinstance(recategorize, NeuralCategorizer)


@pytest.mark.usefixtures("mock_ocr_request")
def test_ocr_request(mocked_responses, api_client):
    ocr = api_client.neural.ocr_request("abc123", [2, 3])
    assert len(mocked_responses.calls) == 1
    request = mocked_responses.calls[0].request
    assert json.loads(request.body) == {"file_id": "abc123", "pages": [2, 3]}
    assert len(ocr.ocr) == 2
    assert ocr.ocr[0] == "This is page 1"
    assert ocr.ocr[1] == "This is page 2"
    assert ocr.processed_pages == 2


@pytest.mark.usefixtures("mock_clean_conversation")
def test_clean_conversation(mocked_responses, api_client):
    convo_response = api_client.neural.clean_conversation(["abc123"])
    assert len(mocked_responses.calls) == 1
    request = mocked_responses.calls[0].request
    assert json.loads(request.body) == {"message_id": ["abc123"]}
    assert len(convo_response) == 1
    convo = convo_response[0]
    assert (
        convo.conversation
        == "<img src='cid:1781777f666586677621' /> This is the conversation"
    )
    assert convo.model_version == "0.0.1"


@pytest.mark.usefixtures("mock_clean_conversation")
def test_clean_conversation_options(mocked_responses, api_client):
    options = NeuralMessageOptions(False, False, False, False, False)
    api_client.neural.clean_conversation(["abc123"], options)
    request = mocked_responses.calls[0].request
    assert json.loads(request.body) == {
        "message_id": ["abc123"],
        "ignore_links": False,
        "ignore_images": False,
        "ignore_tables": False,
        "remove_conclusion_phrases": False,
        "images_as_markdowns": False,
    }


@pytest.mark.usefixtures("mock_clean_conversation")
def test_clean_conversation_extract_images(mocked_responses, api_client):
    convo = api_client.neural.clean_conversation(["abc123"])
    extracted_files = convo[0].extract_images()
    assert len(mocked_responses.calls) == 2
    assert len(extracted_files) == 1
    assert isinstance(extracted_files[0], File) is True
    assert extracted_files[0].id == "1781777f666586677621"



from copy import copy
from nylas.utils import convert_metadata_pairs_to_array

CHUNK_SIZE = 50


class RestfulModelCollection(object):
    def __init__(self, cls, api, filter=None, offset=0, **filters):
        if filter:
            filters.update(filter)
        from nylas.client import APIClient

        if not isinstance(api, APIClient):
            raise Exception("Provided api was not an APIClient.")

        filters.setdefault("offset", offset)

        self.model_class = cls
        self.filters = filters
        self.api = api

    def __iter__(self):
        return self.values()

    def values(self):
        limit = self.filters.get("limit")
        offset = self.filters["offset"]
        fetched = 0
        # Currently, the Nylas API handles pagination poorly: API responses do not expose
        # any information about pagination, so the client does not know whether there is
        # another page of data or not. For example, if the client sends an API request
        # without a limit specified, and the response contains 100 items, how can it tell
        # if there are 100 items in total, or if there more items to fetch on the next page?
        # It can't! The only way to know is to ask for the next page (by repeating the API
        # request with `offset=100`), and see if you get more items or not.
        # If it does not receive more items, it can assume that it has retrieved all the data.
        while True:
            if limit:
                if fetched >= limit:
                    break

                req_limit = min(CHUNK_SIZE, limit - fetched)
            else:
                req_limit = CHUNK_SIZE

            models = self._get_model_collection(offset + fetched, req_limit)
            if not models:
                break

            for model in models:
                yield model

            fetched += len(models)

    def first(self):
        results = self._get_model_collection(0, 1)
        if results:
            return results[0]
        return None

    def all(self, limit=float("infinity")):
        if "limit" in self.filters and self.filters["limit"] is not None:
            limit = self.filters["limit"]
        return self._range(self.filters["offset"], limit)

    def count(self):
        """
        Get the number of objects in the collection being queried

        Returns:
            int: The number of objects in the collection being queried
        """
        self.filters["view"] = "count"
        response = self.api._get_resource_raw(
            self.model_class, resource_id=None, **self.filters
        ).json()
        return response["count"]

    def where(self, filter=None, **filters):
        # Some API parameters like "from" and "in" also are
        # Python reserved keywords. To work around this, we rename
        # them to "from_" and "in_". The API still needs them in
        # their correct form though.
        reserved_keywords = ["from", "in"]
        for keyword in reserved_keywords:
            escaped_keyword = "{}_".format(keyword)
            if escaped_keyword in filters:
                filters[keyword] = filters.get(escaped_keyword)
                del filters[escaped_keyword]

        if filter:
            filters.update(filter)
        filters.setdefault("offset", 0)

        if "metadata_pair" in filters:
            pairs = convert_metadata_pairs_to_array(filters["metadata_pair"])
            filters["metadata_pair"] = pairs

        collection = copy(self)
        collection.filters = filters
        return collection

    def get(self, id):
        return self._get_model(id)

    def create(self, **kwargs):
        return self.model_class.create(self.api, **kwargs)

    def delete(self, id, data=None, **kwargs):
        return self.api._delete_resource(self.model_class, id, data=data, **kwargs)

    def search(self, q, limit=None, offset=None):  # pylint: disable=invalid-name
        from nylas.client.restful_models import (
            Message,
            Thread,
        )  # pylint: disable=cyclic-import

        if self.model_class is Thread or self.model_class is Message:
            kwargs = {"q": q}
            if limit is not None:
                kwargs["limit"] = limit
            if offset is not None:
                kwargs["offset"] = offset
            return self.api._get_resources(self.model_class, extra="search", **kwargs)
        else:
            raise Exception("Searching is only allowed on Thread and Message models")

    def __getitem__(self, key):
        if isinstance(key, slice):
            if key.step is not None:
                raise ValueError(
                    "'step' not supported for slicing "
                    "RestfulModelCollection objects "
                    "(e.g. messages[::step])"
                )
            elif key.start < 0 or key.stop < 0:
                raise ValueError("slice indices must be positive")
            elif key.stop - key.start < 0:
                raise ValueError(
                    "ending slice index cannot be less than " "starting index"
                )
            return self._range(key.start, key.stop - key.start)
        else:
            return self._get_model_collection(key, 1)[0]

    # Private functions

    def _get_model_collection(self, offset=0, limit=CHUNK_SIZE):
        filters = copy(self.filters)
        filters["offset"] = offset
        if not filters.get("limit"):
            filters["limit"] = limit

        return self.api._get_resources(self.model_class, **filters)

    def _get_model(self, id):
        return self.api._get_resource(self.model_class, id, **self.filters)

    def _range(self, offset=0, limit=CHUNK_SIZE):
        accumulated = []
        while len(accumulated) < limit:
            to_fetch = min(limit - len(accumulated), CHUNK_SIZE)
            results = self._get_model_collection(offset + len(accumulated), to_fetch)
            accumulated.extend(results)

            # done if we run out of data to fetch
            if not results or len(results) < to_fetch:
                break

        return accumulated



from nylas.client.restful_models import RestfulModel, Message, File, Contact
from nylas.utils import HttpMethod
import re


def _add_options_to_body(body, options):
    options_dict = options.__dict__
    # Only append set options to body to prevent a 400 error
    options_filtered = {k: v for k, v in options_dict.items() if v is not None}
    return body.update(options_filtered)


class Neural(RestfulModel):
    def __init__(self, api):
        RestfulModel.__init__(self, Neural, api)

    def sentiment_analysis_message(self, message_ids):
        body = {"message_id": message_ids}
        return self.api._request_neural_resource(NeuralSentimentAnalysis, body)

    def sentiment_analysis_text(self, text):
        body = {"text": text}
        return self.api._request_neural_resource(NeuralSentimentAnalysis, body)

    def extract_signature(self, message_ids, parse_contacts=None, options=None):
        body = {"message_id": message_ids}
        if parse_contacts is not None and isinstance(parse_contacts, bool):
            body["parse_contacts"] = parse_contacts
        if options is not None and isinstance(options, NeuralMessageOptions):
            _add_options_to_body(body, options)
        signatures = self.api._request_neural_resource(NeuralSignatureExtraction, body)
        if parse_contacts is not False:
            for sig in signatures:
                sig.contacts = NeuralSignatureContact.create(self.api, **sig.contacts)
        return signatures

    def ocr_request(self, file_id, pages=None):
        body = {"file_id": file_id}
        if pages is not None and isinstance(pages, list):
            body["pages"] = pages
        return self.api._request_neural_resource(NeuralOcr, body)

    def categorize(self, message_ids):
        body = {"message_id": message_ids}
        categorized = self.api._request_neural_resource(NeuralCategorizer, body)
        for message in categorized:
            message.categorizer = Categorize.create(self.api, **message.categorizer)
        return categorized

    def clean_conversation(self, message_ids, options=None):
        body = {"message_id": message_ids}
        if options is not None and isinstance(options, NeuralMessageOptions):
            _add_options_to_body(body, options)
        return self.api._request_neural_resource(NeuralCleanConversation, body)


class NeuralMessageOptions:
    def __init__(
        self,
        ignore_links=None,
        ignore_images=None,
        ignore_tables=None,
        remove_conclusion_phrases=None,
        images_as_markdowns=None,
    ):
        self.ignore_links = ignore_links
        self.ignore_images = ignore_images
        self.ignore_tables = ignore_tables
        self.remove_conclusion_phrases = remove_conclusion_phrases
        self.images_as_markdowns = images_as_markdowns


class NeuralSentimentAnalysis(RestfulModel):
    attrs = [
        "account_id",
        "sentiment",
        "sentiment_score",
        "processed_length",
        "text",
    ]
    collection_name = "sentiment"

    def __init__(self, api):
        RestfulModel.__init__(self, NeuralSentimentAnalysis, api)


class NeuralSignatureExtraction(Message):
    attrs = Message.attrs + ["signature", "model_version", "contacts"]
    collection_name = "signature"

    def __init__(self, api):
        RestfulModel.__init__(self, NeuralSignatureExtraction, api)


class NeuralSignatureContact(RestfulModel):
    attrs = ["job_titles", "links", "phone_numbers", "emails", "names"]
    collection_name = "signature_contact"

    def __init__(self, api):
        RestfulModel.__init__(self, NeuralSignatureContact, api)

    def to_contact_object(self):
        contact = {}
        if self.names is not None:
            contact["given_name"] = self.names[0]["first_name"]
            contact["surname"] = self.names[0]["last_name"]
        if self.job_titles is not None:
            contact["job_title"] = self.job_titles[0]
        if self.emails is not None:
            contact["emails"] = []
            for email in self.emails:
                contact["emails"].append({"type": "personal", "email": email})
        if self.phone_numbers is not None:
            contact["phone_numbers"] = []
            for number in self.phone_numbers:
                contact["phone_numbers"].append({"type": "mobile", "number": number})
        if self.links is not None:
            contact["web_pages"] = []
            for url in self.links:
                description = url["description"] if url["description"] else "homepage"
                contact["web_pages"].append({"type": description, "url": url["url"]})

        return Contact.create(self.api, **contact)


class NeuralCategorizer(Message):
    attrs = Message.attrs + ["categorizer"]
    collection_name = "categorize"

    def __init__(self, api):
        RestfulModel.__init__(self, NeuralCategorizer, api)

    def recategorize(self, category):
        data = {"message_id": self.id, "category": category}
        self.api._request_neural_resource(
            NeuralCategorizer, data, "categorize/feedback", method=HttpMethod.POST
        )
        data = {"message_id": self.id}
        response = self.api._request_neural_resource(NeuralCategorizer, data)
        categorize = response[0]
        if categorize.categorizer:
            categorize.categorizer = Categorize.create(
                self.api, **categorize.categorizer
            )
        return categorize


class Categorize(RestfulModel):
    attrs = ["category", "categorized_at", "model_version", "subcategories"]
    datetime_attrs = {"categorized_at": "categorized_at"}
    collection_name = "category"

    def __init__(self, api):
        RestfulModel.__init__(self, Categorize, api)


class NeuralCleanConversation(Message):
    attrs = Message.attrs + [
        "conversation",
        "model_version",
    ]
    collection_name = "conversation"

    def __init__(self, api):
        RestfulModel.__init__(self, NeuralCleanConversation, api)

    def extract_images(self):
        pattern = "[\(']cid:(.*?)[\)']"
        file_ids = re.findall(pattern, self.conversation)
        files = []
        for match in file_ids:
            files.append(self.api.files.get(match))
        return files


class NeuralOcr(File):
    attrs = File.attrs + [
        "ocr",
        "processed_pages",
    ]
    collection_name = "ocr"

    def __init__(self, api):
        RestfulModel.__init__(self, NeuralOcr, api)



from datetime import datetime
from collections import defaultdict
from enum import Enum

from six import StringIO
from nylas.client.restful_model_collection import RestfulModelCollection
from nylas.client.errors import FileUploadError, UnSyncedError, NylasApiError
from nylas.utils import timestamp_from_dt, AuthMethod

# pylint: disable=attribute-defined-outside-init


def typed_dict_attr(items, attr_name=None):
    if attr_name:
        pairs = [(item["type"], item[attr_name]) for item in items]
    else:
        pairs = [(item["type"], item) for item in items]
    dct = defaultdict(list)
    for key, value in pairs:
        dct[key].append(value)
    return dct


def _is_subclass(cls, parent):
    for base in cls.__bases__:
        if base.__name__.lower() == parent:
            return True
    return False


class RestfulModel(dict):
    attrs = []
    date_attrs = {}
    datetime_attrs = {}
    datetime_filter_attrs = {}
    typed_dict_attrs = {}
    read_only_attrs = {}
    auth_method = AuthMethod.BEARER
    # The Nylas API holds most objects for an account directly under '/',
    # but some of them are under '/a' (mostly the account-management
    # and billing code). api_root is a tiny metaprogramming hack to let
    # us use the same code for both.
    api_root = None

    def __init__(self, cls, api):
        self.id = None
        self.cls = cls
        self.api = api
        super(RestfulModel, self).__init__()

    __setattr__ = dict.__setitem__
    __delattr__ = dict.__delitem__
    __getattr__ = dict.get

    @classmethod
    def create(cls, api, **kwargs):
        object_type = kwargs.get("object")
        cls_object_type = getattr(cls, "object_type", cls.__name__.lower())
        # These are classes that should bypass the check below because they
        # often represent other types (e.g. a delta's object type might be event)
        class_check_whitelist = ["jobstatus", "delta"]
        if (
            object_type
            and object_type != cls_object_type
            and object_type != "account"
            and cls_object_type not in class_check_whitelist
            and not _is_subclass(cls, object_type)
        ):
            # We were given a specific object type and we're trying to
            # instantiate something different; abort. (Relevant for folders
            # and labels API.)
            # We need a special case for accounts because the /accounts API
            # is different between the open source and hosted API.
            # And a special case for job status because the object refers to
            # the type of objects' job status
            return
        obj = cls(api)  # pylint: disable=no-value-for-parameter
        obj.cls = cls
        for attr in cls.attrs:
            # Support attributes we want to override with properties where
            # the property names overlap with the JSON names (e.g. folders)
            attr_name = attr
            if attr_name.startswith("_"):
                attr = attr_name[1:]
            if attr in kwargs:
                obj[attr_name] = kwargs[attr]
                if attr_name == "from":
                    obj["from_"] = kwargs[attr]
        for date_attr, iso_attr in cls.date_attrs.items():
            if kwargs.get(iso_attr):
                obj[date_attr] = datetime.strptime(kwargs[iso_attr], "%Y-%m-%d").date()
        for dt_attr, ts_attr in cls.datetime_attrs.items():
            if kwargs.get(ts_attr):
                try:
                    obj[dt_attr] = datetime.utcfromtimestamp(kwargs[ts_attr])
                except TypeError:
                    # If the datetime format is in the format of ISO8601
                    obj[dt_attr] = datetime.strptime(
                        kwargs[ts_attr], "%Y-%m-%dT%H:%M:%S.%fZ"
                    )
        for attr, value_attr_name in cls.typed_dict_attrs.items():
            obj[attr] = typed_dict_attr(kwargs.get(attr, []), attr_name=value_attr_name)

        if "id" not in kwargs:
            obj["id"] = None

        return obj

    def as_json(self, enforce_read_only=True):
        dct = {}
        # Some API parameters like "from" and "in" also are
        # Python reserved keywords. To work around this, we rename
        # them to "from_" and "in_". The API still needs them in
        # their correct form though.
        reserved_keywords = ["from", "in"]
        for attr in self.cls.attrs:
            if attr in self.read_only_attrs and enforce_read_only is True:
                continue
            if hasattr(self, attr):
                if attr in reserved_keywords:
                    attr_value = getattr(self, "{}_".format(attr))
                else:
                    attr_value = getattr(self, attr)
                if attr_value is not None:
                    dct[attr] = attr_value
        for date_attr, iso_attr in self.cls.date_attrs.items():
            if date_attr in self.read_only_attrs and enforce_read_only is True:
                continue
            if self.get(date_attr):
                dct[iso_attr] = self[date_attr].strftime("%Y-%m-%d")
        for dt_attr, ts_attr in self.cls.datetime_attrs.items():
            if dt_attr in self.read_only_attrs and enforce_read_only is True:
                continue
            if self.get(dt_attr):
                dct[ts_attr] = timestamp_from_dt(self[dt_attr])
        for attr, value_attr in self.cls.typed_dict_attrs.items():
            if attr in self.read_only_attrs and enforce_read_only is True:
                continue
            typed_dict = getattr(self, attr)
            if value_attr:
                dct[attr] = []
                for key, values in typed_dict.items():
                    for value in values:
                        dct[attr].append({"type": key, value_attr: value})
            else:
                dct[attr] = []
                for values in typed_dict.values():
                    for value in values:
                        dct[attr].append(value)
        return dct


class NylasAPIObject(RestfulModel):
    read_only_attrs = {"id", "account_id", "object", "job_status_id"}

    def __init__(self, cls, api):
        RestfulModel.__init__(self, cls, api)

    def child_collection(self, cls, **filters):
        return RestfulModelCollection(cls, self.api, **filters)

    def save(self, **kwargs):
        if self.id:
            new_obj = self._update_resource(**kwargs)
        else:
            new_obj = self._create_resource(**kwargs)
        self._update_values(new_obj)

    def update(self):
        new_obj = self._update_resource()
        self._update_values(new_obj)

    def _create_resource(self, **kwargs):
        return self.api._create_resource(self.cls, self.as_json(), **kwargs)

    def _update_resource(self, **kwargs):
        return self.api._update_resource(self.cls, self.id, self.as_json(), **kwargs)

    def _update_values(self, new_obj):
        for attr in self.cls.attrs:
            if hasattr(new_obj, attr):
                setattr(self, attr, getattr(new_obj, attr))


class Message(NylasAPIObject):
    attrs = [
        "bcc",
        "body",
        "cc",
        "date",
        "events",
        "files",
        "from",
        "id",
        "account_id",
        "object",
        "snippet",
        "starred",
        "subject",
        "thread_id",
        "job_status_id",
        "to",
        "unread",
        "starred",
        "metadata",
        "_folder",
        "_labels",
        "headers",
        "reply_to",
    ]
    datetime_attrs = {"received_at": "date"}
    datetime_filter_attrs = {
        "received_before": "received_before",
        "received_after": "received_after",
    }
    collection_name = "messages"

    def __init__(self, api):
        NylasAPIObject.__init__(self, Message, api)

    @property
    def attachments(self):
        return self.child_collection(File, message_id=self.id)

    @property
    def folder(self):
        # Instantiate a Folder object from the API response
        if self._folder:
            return Folder.create(self.api, **self._folder)

    @property
    def labels(self):
        if self._labels:
            return [Label.create(self.api, **l) for l in self._labels]
        return []

    def update_folder(self, folder_id):
        update = {"folder": folder_id}
        new_obj = self.api._update_resource(self.cls, self.id, update)
        for attr in self.cls.attrs:
            if hasattr(new_obj, attr):
                setattr(self, attr, getattr(new_obj, attr))
        return self.folder

    def update_labels(self, label_ids=None):
        label_ids = label_ids or []
        update = {"labels": label_ids}
        new_obj = self.api._update_resource(self.cls, self.id, update)
        for attr in self.cls.attrs:
            if hasattr(new_obj, attr):
                setattr(self, attr, getattr(new_obj, attr))
        return self.labels

    def add_labels(self, label_ids=None):
        label_ids = label_ids or []
        labels = [l.id for l in self.labels]
        labels = list(set(labels).union(set(label_ids)))
        return self.update_labels(labels)

    def add_label(self, label_id):
        return self.add_labels([label_id])

    def remove_labels(self, label_ids=None):
        label_ids = label_ids or []
        labels = [l.id for l in self.labels]
        labels = list(set(labels) - set(label_ids))
        return self.update_labels(labels)

    def remove_label(self, label_id):
        return self.remove_labels([label_id])

    def mark_as_seen(self):
        self.mark_as_read()

    def mark_as_read(self):
        update = {"unread": False}
        self.api._update_resource(self.cls, self.id, update)
        self.unread = False

    def mark_as_unread(self):
        update = {"unread": True}
        self.api._update_resource(self.cls, self.id, update)
        self.unread = True

    def star(self):
        update = {"starred": True}
        self.api._update_resource(self.cls, self.id, update)
        self.starred = True

    def unstar(self):
        update = {"starred": False}
        self.api._update_resource(self.cls, self.id, update)
        self.starred = False

    @property
    def raw(self):
        headers = {"Accept": "message/rfc822"}
        response = self.api._get_resource_raw(Message, self.id, headers=headers)
        if response.status_code == 202:
            raise UnSyncedError(response.content)
        return response.content


class Folder(NylasAPIObject):
    attrs = ["id", "display_name", "name", "object", "account_id", "job_status_id"]
    collection_name = "folders"

    def __init__(self, api):
        NylasAPIObject.__init__(self, Folder, api)

    @property
    def threads(self):
        return self.child_collection(Thread, folder_id=self.id)

    @property
    def messages(self):
        return self.child_collection(Message, folder_id=self.id)


class Label(NylasAPIObject):
    attrs = ["id", "display_name", "name", "object", "account_id", "job_status_id"]
    collection_name = "labels"

    def __init__(self, api):
        NylasAPIObject.__init__(self, Label, api)

    @property
    def threads(self):
        return self.child_collection(Thread, label_id=self.id)

    @property
    def messages(self):
        return self.child_collection(Message, label_id=self.id)


class Thread(NylasAPIObject):
    attrs = [
        "draft_ids",
        "id",
        "message_ids",
        "account_id",
        "object",
        "participants",
        "snippet",
        "subject",
        "subject_date",
        "last_message_timestamp",
        "first_message_timestamp",
        "last_message_received_timestamp",
        "last_message_sent_timestamp",
        "unread",
        "starred",
        "version",
        "_folders",
        "_labels",
        "received_recent_date",
        "has_attachments",
    ]
    datetime_attrs = {
        "first_message_at": "first_message_timestamp",
        "last_message_at": "last_message_timestamp",
        "last_message_received_at": "last_message_received_timestamp",
        "last_message_sent_at": "last_message_sent_timestamp",
    }
    datetime_filter_attrs = {
        "last_message_before": "last_message_before",
        "last_message_after": "last_message_after",
        "started_before": "started_before",
        "started_after": "started_after",
    }
    collection_name = "threads"

    def __init__(self, api):
        NylasAPIObject.__init__(self, Thread, api)

    @property
    def messages(self):
        return self.child_collection(Message, thread_id=self.id)

    @property
    def drafts(self):
        return self.child_collection(Draft, thread_id=self.id)

    @property
    def folders(self):
        if self._folders:
            return [Folder.create(self.api, **f) for f in self._folders]
        return []

    @property
    def labels(self):
        if self._labels:
            return [Label.create(self.api, **l) for l in self._labels]
        return []

    def update_folder(self, folder_id):
        update = {"folder": folder_id}
        new_obj = self.api._update_resource(self.cls, self.id, update)
        for attr in self.cls.attrs:
            if hasattr(new_obj, attr):
                setattr(self, attr, getattr(new_obj, attr))
        return self.folder

    def update_labels(self, label_ids=None):
        label_ids = label_ids or []
        update = {"labels": label_ids}
        new_obj = self.api._update_resource(self.cls, self.id, update)
        for attr in self.cls.attrs:
            if hasattr(new_obj, attr):
                setattr(self, attr, getattr(new_obj, attr))
        return self.labels

    def add_labels(self, label_ids=None):
        label_ids = label_ids or []
        labels = [l.id for l in self.labels]
        labels = list(set(labels).union(set(label_ids)))
        return self.update_labels(labels)

    def add_label(self, label_id):
        return self.add_labels([label_id])

    def remove_labels(self, label_ids=None):
        label_ids = label_ids or []
        labels = [l.id for l in self.labels]
        labels = list(set(labels) - set(label_ids))
        return self.update_labels(labels)

    def remove_label(self, label_id):
        return self.remove_labels([label_id])

    def mark_as_seen(self):
        self.mark_as_read()

    def mark_as_read(self):
        update = {"unread": False}
        self.api._update_resource(self.cls, self.id, update)
        self.unread = False

    def mark_as_unread(self):
        update = {"unread": True}
        self.api._update_resource(self.cls, self.id, update)
        self.unread = True

    def star(self):
        update = {"starred": True}
        self.api._update_resource(self.cls, self.id, update)
        self.starred = True

    def unstar(self):
        update = {"starred": False}
        self.api._update_resource(self.cls, self.id, update)
        self.starred = False

    def create_reply(self):
        draft = self.drafts.create()
        draft.thread_id = self.id
        draft.subject = self.subject
        return draft


# This is a dummy class that allows us to use the create_resource function
# and pass in a 'Send' object that will translate into a 'send' endpoint.
class Send(Message):
    collection_name = "send"

    def __init__(self, api):  # pylint: disable=super-init-not-called
        NylasAPIObject.__init__(
            self, Send, api
        )  # pylint: disable=non-parent-init-called


class Draft(Message):
    attrs = [
        "bcc",
        "cc",
        "body",
        "date",
        "files",
        "from",
        "id",
        "account_id",
        "object",
        "subject",
        "thread_id",
        "to",
        "job_status_id",
        "unread",
        "version",
        "file_ids",
        "reply_to_message_id",
        "reply_to",
        "starred",
        "snippet",
        "tracking",
        "metadata",
    ]
    datetime_attrs = {"last_modified_at": "date"}
    collection_name = "drafts"

    def __init__(self, api, thread_id=None):  # pylint: disable=unused-argument
        Message.__init__(self, api)
        NylasAPIObject.__init__(
            self, Thread, api
        )  # pylint: disable=non-parent-init-called
        self.file_ids = []

    def attach(self, file):
        if not file.id:
            file.save()

        self.file_ids.append(file.id)

    def detach(self, file):
        if file.id in self.file_ids:
            self.file_ids.remove(file.id)

    def send(self):
        if not self.id:
            data = self.as_json()
        else:
            data = {"draft_id": self.id}
            if hasattr(self, "version"):
                data["version"] = self.version
            if hasattr(self, "tracking") and self.tracking is not None:
                data["tracking"] = self.tracking

        msg = self.api._create_resource(Send, data)
        if msg:
            return msg

    def delete(self):
        if self.id and self.version is not None:
            data = {"version": self.version}
            self.api._delete_resource(self.cls, self.id, data=data)


class File(NylasAPIObject):
    attrs = [
        "content_type",
        "filename",
        "id",
        "content_id",
        "account_id",
        "object",
        "size",
        "message_ids",
    ]
    collection_name = "files"

    def save(self):  # pylint: disable=arguments-differ
        stream = getattr(self, "stream", None)
        if not stream:
            data = getattr(self, "data", None)
            if data:
                stream = StringIO(data)

        if not stream:
            message = (
                "File object not properly formatted, "
                "must provide either a stream or data."
            )
            raise FileUploadError(message)

        file_info = (self.filename, stream, self.content_type, {})  # upload headers

        new_obj = self.api._create_resources(File, {"file": file_info})
        new_obj = new_obj[0]
        for attr in self.attrs:
            if hasattr(new_obj, attr):
                setattr(self, attr, getattr(new_obj, attr))

    def download(self):
        if not self.id:
            message = "Can't download a file that hasn't been uploaded."
            raise FileUploadError(message)

        return self.api._get_resource_data(File, self.id, extra="download")

    def __init__(self, api):
        NylasAPIObject.__init__(self, File, api)


class Contact(NylasAPIObject):
    attrs = [
        "id",
        "object",
        "account_id",
        "given_name",
        "middle_name",
        "surname",
        "suffix",
        "nickname",
        "company_name",
        "job_title",
        "job_status_id",
        "manager_name",
        "office_location",
        "source",
        "notes",
        "picture_url",
    ]
    date_attrs = {"birthday": "birthday"}
    typed_dict_attrs = {
        "emails": "email",
        "im_addresses": "im_address",
        "physical_addresses": None,
        "phone_numbers": "number",
        "web_pages": "url",
    }
    collection_name = "contacts"

    def __init__(self, api):
        NylasAPIObject.__init__(self, Contact, api)

    def get_picture(self):
        if not self.get("picture_url", None):
            return None

        response = self.api._get_resource_raw(
            Contact, self.id, extra="picture", stream=True
        )
        if response.status_code >= 400:
            raise NylasApiError(response)
        return response.raw


class Calendar(NylasAPIObject):
    attrs = [
        "id",
        "account_id",
        "name",
        "description",
        "hex_color",
        "job_status_id",
        "metadata",
        "read_only",
        "is_primary",
        "object",
    ]
    collection_name = "calendars"

    def __init__(self, api):
        NylasAPIObject.__init__(self, Calendar, api)
        self.read_only_attrs.update({"is_primary", "read_only", "hex_color"})

    @property
    def events(self):
        return self.child_collection(Event, calendar_id=self.id)


class Event(NylasAPIObject):
    attrs = [
        "id",
        "account_id",
        "title",
        "description",
        "conferencing",
        "location",
        "read_only",
        "when",
        "busy",
        "participants",
        "calendar_id",
        "recurrence",
        "status",
        "master_event_id",
        "job_status_id",
        "owner",
        "original_start_time",
        "object",
        "message_id",
        "ical_uid",
        "metadata",
        "notifications",
        "event_collection_id",
        "capacity",
        "round_robin_order",
        "visibility",
    ]
    datetime_attrs = {"original_start_at": "original_start_time"}
    collection_name = "events"

    def __init__(self, api):
        NylasAPIObject.__init__(self, Event, api)
        self.read_only_attrs.update(
            {
                "ical_uid",
                "message_id",
                "owner",
                "status",
                "master_event_id",
                "original_start_time",
            }
        )

    def as_json(self, enforce_read_only=True):
        dct = NylasAPIObject.as_json(self, enforce_read_only)
        if enforce_read_only is False:
            return dct

        # Filter some parameters we got from the API
        if dct.get("when"):
            # Currently, the event (self) and the dict (dct) share the same
            # reference to the `'when'` dict.  We need to clone the dict so
            # that when we remove the object key, the original event's
            # `'when'` reference is unmodified.
            dct["when"] = dct["when"].copy()
            dct["when"].pop("object", None)

        if dct.get("participants") and isinstance(dct.get("participants"), list):
            # The status of a participant cannot be updated and, if the key is
            # included, it will return an error from the API
            for participant in dct.get("participants"):
                participant.pop("status", None)

        return dct

    def rsvp(self, status, comment=None):
        if not self.message_id:
            raise ValueError(
                "This event was not imported from an iCalendar invite, and so it is not possible to RSVP via Nylas"
            )
        if status not in {"yes", "no", "maybe"}:
            raise ValueError("invalid status: {status}".format(status=status))

        url = "{api_server}/send-rsvp".format(api_server=self.api.api_server)
        data = {
            "event_id": self.id,
            "status": status,
            "comment": comment,
        }
        response = self.api.session.post(url, json=data)
        if response.status_code >= 400:
            raise NylasApiError(response)
        result = response.json()
        return Event.create(self, **result)

    def generate_ics(self, ical_uid=None, method=None, prodid=None):
        """
        Generate an ICS file server-side, from an Event

        Args:
            ical_uid (str): Unique identifier used events across calendaring systems
            method (str): Description of invitation and response methods for attendees
            prodid (str): Company-specific unique product identifier

        Returns:
            str: String for writing directly into an ICS file

        Raises:
            ValueError: If the event does not have calendar_id or when set
            RuntimeError: If the server returns an object without an ics string
        """
        if not self.calendar_id or not self.when:
            raise ValueError(
                "Cannot generate an ICS file for an event without a Calendar ID or when set"
            )

        payload = {}
        ics_options = {}
        if self.id:
            payload["event_id"] = self.id
        else:
            payload = self.as_json()

        if ical_uid:
            ics_options["ical_uid"] = ical_uid
        if method:
            ics_options["method"] = method
        if prodid:
            ics_options["prodid"] = prodid

        if ics_options:
            payload["ics_options"] = ics_options

        response = self.api._post_resource(Event, None, "to-ics", payload)
        if "ics" in response:
            return response["ics"]
        raise RuntimeError(
            "Unexpected response from the API server. Returned 200 but no 'ics' string found."
        )

    def validate(self):
        if (
            self.conferencing
            and "details" in self.conferencing
            and "autocreate" in self.conferencing
        ):
            raise ValueError(
                "Cannot set both 'details' and 'autocreate' in conferencing object."
            )
        if (
            self.capacity
            and self.capacity != -1
            and self.participants
            and len(self.participants) > self.capacity
        ):
            raise ValueError(
                "The number of participants in the event exceeds the set capacity."
            )

    def save(self, **kwargs):
        self.validate()

        super(Event, self).save(**kwargs)


class RoomResource(NylasAPIObject):
    attrs = [
        "object",
        "email",
        "name",
        "capacity",
        "building",
        "floor_name",
        "floor_number",
    ]
    object_type = "room_resource"
    collection_name = "resources"

    def __init__(self, api):
        NylasAPIObject.__init__(self, RoomResource, api)


class JobStatus(NylasAPIObject):
    attrs = [
        "id",
        "account_id",
        "job_status_id",
        "action",
        "object",
        "status",
        "original_data",
        "metadata",
    ]
    datetime_attrs = {"created_at": "created_at"}
    collection_name = "job-statuses"

    def __init__(self, api):
        NylasAPIObject.__init__(self, JobStatus, api)
        self.read_only_attrs.update(
            {
                "action",
                "status",
                "original_data",
            }
        )

    def is_successful(self):
        return self.status == "successful"


class Scheduler(NylasAPIObject):
    attrs = [
        "id",
        "access_tokens",
        "app_client_id",
        "app_organization_id",
        "config",
        "edit_token",
        "name",
        "slug",
    ]
    date_attrs = {
        "created_at": "created_at",
        "modified_at": "modified_at",
    }
    collection_name = "manage/pages"

    def __init__(self, api):
        NylasAPIObject.__init__(self, Scheduler, api)

    def get_available_calendars(self):
        if not self.id:
            raise ValueError("Cannot get calendars for a page without an ID")

        response = self.api._get_resource_raw(Scheduler, self.id, extra="calendars")
        response_body = response.json()
        for body in response_body:
            for i in range(len(body["calendars"])):
                body["calendars"][i] = Calendar.create(self.api, **body["calendars"][i])

        return response_body

    def upload_image(self, content_type, object_name):
        if not self.id:
            raise ValueError("Cannot upload an image to a page without an ID")

        data = {"contentType": content_type, "objectName": object_name}
        response = self.api._put_resource(
            Scheduler, self.id, data, extra="upload-image"
        )
        return response


class Component(NylasAPIObject):
    attrs = [
        "id",
        "account_id",
        "name",
        "type",
        "action",
        "active",
        "settings",
        "public_account_id",
        "public_token_id",
        "public_application_id",
        "access_token",
        "allowed_domains",
    ]
    datetime_attrs = {
        "created_at": "created_at",
        "updated_at": "updated_at",
    }

    collection_name = None
    api_root = "component"

    def __init__(self, api):
        NylasAPIObject.__init__(self, Component, api)
        self.read_only_attrs.update(
            {
                "public_application_id",
                "created_at",
                "updated_at",
            }
        )

    def as_json(self, enforce_read_only=True):
        dct = NylasAPIObject.as_json(self, enforce_read_only)
        if enforce_read_only is False:
            return dct

        # "type" cannot be modified after created
        if self.id:
            dct.pop("type")
        return dct


class Webhook(NylasAPIObject):
    attrs = (
        "id",
        "callback_url",
        "state",
        "triggers",
        "application_id",
        "version",
    )

    collection_name = "webhooks"
    api_root = "a"

    def __init__(self, api):
        NylasAPIObject.__init__(self, Webhook, api)
        self.read_only_attrs.update({"application_id", "version"})

    def as_json(self, enforce_read_only=True):
        dct = {}
        # Only 'state' can get updated
        if self.id and enforce_read_only is True:
            dct["state"] = self.state
        else:
            dct = NylasAPIObject.as_json(self, enforce_read_only)
        return dct

    class Trigger(str, Enum):
        """
        This is an Enum representing all the possible webhook triggers

        see more: https://developer.nylas.com/docs/developer-tools/webhooks/available-webhooks
        """

        ACCOUNT_CONNECTED = "account.connected"
        ACCOUNT_RUNNING = "account.running"
        ACCOUNT_STOPPED = "account.stopped"
        ACCOUNT_INVALID = "account.invalid"
        ACCOUNT_SYNC_ERROR = "account.sync_error"
        MESSAGE_CREATED = "message.created"
        MESSAGE_OPENED = "message.opened"
        MESSAGE_UPDATED = "message.updated"
        MESSAGE_LINK_CLICKED = "message.link_clicked"
        THREAD_REPLIED = "thread.replied"
        CONTACT_CREATED = "contact.created"
        CONTACT_UPDATED = "contact.updated"
        CONTACT_DELETED = "contact.deleted"
        CALENDAR_CREATED = "calendar.created"
        CALENDAR_UPDATED = "calendar.updated"
        CALENDAR_DELETED = "calendar.deleted"
        EVENT_CREATED = "event.created"
        EVENT_UPDATED = "event.updated"
        EVENT_DELETED = "event.deleted"
        JOB_SUCCESSFUL = "job.successful"
        JOB_FAILED = "job.failed"

    class State(str, Enum):
        """
        This is an Enum representing all the possible webhook states

        see more: https://developer.nylas.com/docs/developer-tools/webhooks/#enable-and-disable-webhooks
        """

        ACTIVE = "active"
        INACTIVE = "inactive"


class Namespace(NylasAPIObject):
    attrs = [
        "account",
        "email_address",
        "id",
        "account_id",
        "object",
        "provider",
        "name",
        "organization_unit",
    ]
    collection_name = "n"

    def __init__(self, api):
        NylasAPIObject.__init__(self, Namespace, api)

    def child_collection(self, cls, **filters):
        return RestfulModelCollection(cls, self.api, self.id, **filters)


class Account(NylasAPIObject):
    api_root = "a"

    attrs = [
        "account_id",
        "billing_state",
        "email",
        "id",
        "namespace_id",
        "provider",
        "sync_state",
        "authentication_type",
        "trial",
        "metadata",
    ]

    collection_name = "accounts"

    def __init__(self, api):
        NylasAPIObject.__init__(self, Account, api)

    def as_json(self, enforce_read_only=True):
        if enforce_read_only is False:
            return NylasAPIObject.as_json(self, enforce_read_only)
        else:
            return {"metadata": self.metadata}

    def upgrade(self):
        return self.api._call_resource_method(self, self.account_id, "upgrade", None)

    def downgrade(self):
        return self.api._call_resource_method(self, self.account_id, "downgrade", None)


class APIAccount(NylasAPIObject):
    attrs = [
        "account_id",
        "email_address",
        "id",
        "name",
        "object",
        "organization_unit",
        "provider",
        "sync_state",
    ]
    datetime_attrs = {"linked_at": "linked_at"}

    collection_name = "accounts"

    def __init__(self, api):
        NylasAPIObject.__init__(self, APIAccount, api)


class SingletonAccount(APIAccount):
    # This is an APIAccount that lives under /account.
    collection_name = "account"



import pytest
from nylas.client.restful_models import Label, Thread, Message


@pytest.mark.usefixtures("mock_labels")
def test_list_labels(api_client):
    labels = api_client.labels
    labels = [l for l in labels]
    assert len(labels) == 5
    assert all(isinstance(x, Label) for x in labels)


@pytest.mark.usefixtures("mock_label")
def test_get_label(api_client):
    label = api_client.labels.get("anuep8pe5ugmxrucchrzba2o8")
    assert label is not None
    assert isinstance(label, Label)
    assert label.display_name == "Important"


@pytest.mark.usefixtures("mock_label", "mock_threads")
def test_label_threads(api_client):
    label = api_client.labels.get("anuep8pe5ugmxrucchrzba2o8")
    assert label.threads
    assert all(isinstance(thread, Thread) for thread in label.threads)


@pytest.mark.usefixtures("mock_label", "mock_messages")
def test_label_messages(api_client):
    label = api_client.labels.get("anuep8pe5ugmxrucchrzba2o8")
    assert label.messages
    assert all(isinstance(message, Message) for message in label.messages)



import os
import re
import json
import copy
import cgi
import random
import string
import pytest
import responses
from urlobject import URLObject
from nylas import APIClient

# pylint: disable=redefined-outer-name,too-many-lines

#### HANDLING PAGINATION ####
# Currently, the Nylas API handles pagination poorly: API responses do not expose
# any information about pagination, so the client does not know whether there is
# another page of data or not. For example, if the client sends an API request
# without a limit specified, and the response contains 100 items, how can it tell
# if there are 100 items in total, or if there more items to fetch on the next page?
# It can't! The only way to know is to ask for the next page (by repeating the API
# request with `offset=100`), and see if you get more items or not.
# If it does not receive more items, it can assume that it has retrieved all the data.
#
# This file contains mocks for several API endpoints, including "list" endpoints
# like `/messages` and `/events`. The mocks for these list endpoints must be smart
# enough to check for an `offset` query param, and return an empty list if the
# client requests more data than the first page. If the mock does not
# check for this `offset` query param, and returns the same mock data over and over,
# any SDK method that tries to fetch *all* of a certain type of data
# (like `client.messages.all()`) will never complete.


def generate_id(size=25, chars=string.ascii_letters + string.digits):
    return "".join(random.choice(chars) for _ in range(size))


@pytest.fixture
def message_body():
    return {
        "busy": True,
        "calendar_id": "94rssh7bd3rmsxsp19kiocxze",
        "description": None,
        "id": "cv4ei7syx10uvsxbs21ccsezf",
        "location": "1 Infinite loop, Cupertino",
        "message_id": None,
        "namespace_id": "384uhp3aj8l7rpmv9s2y2rukn",
        "object": "event",
        "owner": None,
        "participants": [],
        "read_only": False,
        "status": "confirmed",
        "title": "The rain song",
        "when": {
            "end_time": 1441056790,
            "object": "timespan",
            "start_time": 1441053190,
        },
    }


@pytest.fixture
def access_token():
    return "l3m0n_w4ter"


@pytest.fixture
def account_id():
    return "4ennivvrcgsqytgybfk912dto"


@pytest.fixture
def api_url():
    return "https://localhost:2222"


@pytest.fixture
def client_id():
    return "fake-client-id"


@pytest.fixture
def client_secret():
    return "nyl4n4ut"


@pytest.fixture
def api_client(api_url):
    return APIClient(
        client_id=None, client_secret=None, access_token=None, api_server=api_url
    )


@pytest.fixture
def api_client_with_client_id(access_token, api_url, client_id, client_secret):
    return APIClient(
        client_id=client_id,
        client_secret=client_secret,
        access_token=access_token,
        api_server=api_url,
    )


@pytest.fixture
def mocked_responses():
    rmock = responses.RequestsMock(assert_all_requests_are_fired=False)
    with rmock:
        yield rmock


@pytest.fixture
def mock_save_draft(mocked_responses, api_url):
    save_endpoint = re.compile(api_url + "/drafts")
    response_body = json.dumps(
        {"id": "4dl0ni6vxomazo73r5oydo16k", "version": "4dw0ni6txomazo33r5ozdo16j"}
    )
    mocked_responses.add(
        responses.POST,
        save_endpoint,
        content_type="application/json",
        status=200,
        body=response_body,
        match_querystring=True,
    )


@pytest.fixture
def mock_account(mocked_responses, api_url, account_id):
    response_body = json.dumps(
        {
            "account_id": account_id,
            "email_address": "ben.bitdiddle1861@gmail.com",
            "id": account_id,
            "name": "Ben Bitdiddle",
            "object": "account",
            "provider": "gmail",
            "organization_unit": "label",
            "billing_state": "paid",
            "linked_at": 1500920299,
            "sync_state": "running",
        }
    )
    mocked_responses.add(
        responses.GET,
        re.compile(api_url + "/account(?!s)/?"),
        content_type="application/json",
        status=200,
        body=response_body,
    )


@pytest.fixture
def mock_accounts(mocked_responses, api_url, account_id, client_id):
    accounts = [
        {
            "account_id": account_id,
            "email_address": "ben.bitdiddle1861@gmail.com",
            "id": account_id,
            "name": "Ben Bitdiddle",
            "object": "account",
            "provider": "gmail",
            "organization_unit": "label",
            "billing_state": "paid",
            "linked_at": 1500920299,
            "sync_state": "running",
        }
    ]

    def list_callback(request):
        url = URLObject(request.url)
        offset = int(url.query_dict.get("offset") or 0)
        if offset:
            return (200, {}, json.dumps([]))
        return (200, {}, json.dumps(accounts))

    def update_callback(request):
        response = accounts[0]
        payload = json.loads(request.body)
        if payload["metadata"]:
            response["metadata"] = payload["metadata"]
        return 200, {}, json.dumps(response)

    def delete_callback(request):
        response = {"success": True}
        return 200, {}, json.dumps(response)

    url_re = "{base}(/a/{client_id})?/accounts/?".format(
        base=api_url, client_id=client_id
    )
    mocked_responses.add_callback(
        responses.GET,
        re.compile(url_re),
        content_type="application/json",
        callback=list_callback,
    )
    mocked_responses.add_callback(
        responses.PUT,
        re.compile(url_re),
        content_type="application/json",
        callback=update_callback,
    )
    mocked_responses.add_callback(
        responses.DELETE,
        re.compile(url_re),
        content_type="application/json",
        callback=delete_callback,
    )


@pytest.fixture
def mock_folder_account(mocked_responses, api_url, account_id):
    response_body = json.dumps(
        {
            "email_address": "ben.bitdiddle1861@office365.com",
            "id": account_id,
            "name": "Ben Bitdiddle",
            "account_id": account_id,
            "object": "account",
            "provider": "eas",
            "organization_unit": "folder",
        }
    )
    mocked_responses.add(
        responses.GET,
        api_url + "/account",
        content_type="application/json",
        status=200,
        body=response_body,
        match_querystring=True,
    )


@pytest.fixture
def mock_labels(mocked_responses, api_url, account_id):
    labels = [
        {
            "display_name": "Important",
            "id": "anuep8pe5ugmxrucchrzba2o8",
            "name": "important",
            "account_id": account_id,
            "object": "label",
        },
        {
            "display_name": "Trash",
            "id": "f1xgowbgcehk235xiy3c3ek42",
            "name": "trash",
            "account_id": account_id,
            "object": "label",
        },
        {
            "display_name": "Sent Mail",
            "id": "ah14wp5fvypvjjnplh7nxgb4h",
            "name": "sent",
            "account_id": account_id,
            "object": "label",
        },
        {
            "display_name": "All Mail",
            "id": "ah14wp5fvypvjjnplh7nxgb4h",
            "name": "all",
            "account_id": account_id,
            "object": "label",
        },
        {
            "display_name": "Inbox",
            "id": "dc11kl3s9lj4760g6zb36spms",
            "name": "inbox",
            "account_id": account_id,
            "object": "label",
        },
    ]

    def list_callback(request):
        url = URLObject(request.url)
        offset = int(url.query_dict.get("offset") or 0)
        if offset:
            return (200, {}, json.dumps([]))
        return (200, {}, json.dumps(labels))

    endpoint = re.compile(api_url + "/labels.*")
    mocked_responses.add_callback(
        responses.GET,
        endpoint,
        content_type="application/json",
        callback=list_callback,
    )


@pytest.fixture
def mock_label(mocked_responses, api_url, account_id):
    response_body = json.dumps(
        {
            "display_name": "Important",
            "id": "anuep8pe5ugmxrucchrzba2o8",
            "name": "important",
            "account_id": account_id,
            "object": "label",
        }
    )
    url = api_url + "/labels/anuep8pe5ugmxrucchrzba2o8"
    mocked_responses.add(
        responses.GET,
        url,
        content_type="application/json",
        status=200,
        body=response_body,
    )


@pytest.fixture
def mock_folder(mocked_responses, api_url, account_id):
    folder = {
        "display_name": "My Folder",
        "id": "anuep8pe5ug3xrupchwzba2o8",
        "name": None,
        "account_id": account_id,
        "object": "folder",
    }
    response_body = json.dumps(folder)
    url = api_url + "/folders/anuep8pe5ug3xrupchwzba2o8"
    mocked_responses.add(
        responses.GET,
        url,
        content_type="application/json",
        status=200,
        body=response_body,
    )

    def request_callback(request):
        payload = json.loads(request.body)
        if "display_name" in payload:
            folder.update(payload)
        return (200, {}, json.dumps(folder))

    def delete_callback(request):
        payload = {"successful": True}
        return 200, {}, json.dumps(payload)

    mocked_responses.add_callback(
        responses.PUT, url, content_type="application/json", callback=request_callback
    )

    mocked_responses.add_callback(
        responses.DELETE, url, content_type="application/json", callback=delete_callback
    )


@pytest.fixture
def mock_messages(mocked_responses, api_url, account_id):
    messages = [
        {
            "id": "1234",
            "to": [{"email": "foo@yahoo.com", "name": "Foo"}],
            "from": [{"email": "bar@gmail.com", "name": "Bar"}],
            "subject": "Test Message",
            "account_id": account_id,
            "object": "message",
            "labels": [{"name": "inbox", "display_name": "Inbox", "id": "abcd"}],
            "starred": False,
            "unread": True,
            "date": 1265077342,
        },
        {
            "id": "1238",
            "to": [{"email": "foo2@yahoo.com", "name": "Foo Two"}],
            "from": [{"email": "bar2@gmail.com", "name": "Bar Two"}],
            "subject": "Test Message 2",
            "account_id": account_id,
            "object": "message",
            "labels": [{"name": "inbox", "display_name": "Inbox", "id": "abcd"}],
            "starred": False,
            "unread": True,
            "date": 1265085342,
        },
        {
            "id": "12",
            "to": [{"email": "foo3@yahoo.com", "name": "Foo Three"}],
            "from": [{"email": "bar3@gmail.com", "name": "Bar Three"}],
            "subject": "Test Message 3",
            "account_id": account_id,
            "object": "message",
            "labels": [{"name": "archive", "display_name": "Archive", "id": "gone"}],
            "starred": False,
            "unread": False,
            "date": 1265093842,
        },
    ]

    def list_callback(request):
        url = URLObject(request.url)
        offset = int(url.query_dict.get("offset") or 0)
        if offset:
            return (200, {}, json.dumps([]))
        return (200, {}, json.dumps(messages))

    endpoint = re.compile(api_url + "/messages")
    mocked_responses.add_callback(
        responses.GET, endpoint, content_type="application/json", callback=list_callback
    )


@pytest.fixture
def mock_message(mocked_responses, api_url, account_id):
    base_msg = {
        "id": "1234",
        "to": [{"email": "foo@yahoo.com", "name": "Foo"}],
        "from": [{"email": "bar@gmail.com", "name": "Bar"}],
        "subject": "Test Message",
        "account_id": account_id,
        "object": "message",
        "labels": [{"name": "inbox", "display_name": "Inbox", "id": "abcd"}],
        "starred": False,
        "unread": True,
    }
    response_body = json.dumps(base_msg)

    def request_callback(request):
        payload = json.loads(request.body)
        if "labels" in payload:
            labels = [
                {"name": "test", "display_name": "test", "id": l}
                for l in payload["labels"]
            ]
            base_msg["labels"] = labels
        if "metadata" in payload:
            base_msg["metadata"] = payload["metadata"]
        return (200, {}, json.dumps(base_msg))

    endpoint = re.compile(api_url + "/messages/1234")
    mocked_responses.add(
        responses.GET,
        endpoint,
        content_type="application/json",
        status=200,
        body=response_body,
    )
    mocked_responses.add_callback(
        responses.PUT,
        endpoint,
        content_type="application/json",
        callback=request_callback,
    )
    mocked_responses.add(
        responses.DELETE, endpoint, content_type="application/json", status=200, body=""
    )


@pytest.fixture
def mock_threads(mocked_responses, api_url, account_id):
    threads = [
        {
            "id": "5678",
            "subject": "Test Thread",
            "account_id": account_id,
            "object": "thread",
            "folders": [{"name": "inbox", "display_name": "Inbox", "id": "abcd"}],
            "starred": True,
            "unread": False,
            "first_message_timestamp": 1451703845,
            "last_message_timestamp": 1483326245,
            "last_message_received_timestamp": 1483326245,
            "last_message_sent_timestamp": 1483232461,
        }
    ]

    def list_callback(request):
        url = URLObject(request.url)
        offset = int(url.query_dict.get("offset") or 0)
        if offset:
            return (200, {}, json.dumps([]))
        return (200, {}, json.dumps(threads))

    endpoint = re.compile(api_url + "/threads")
    mocked_responses.add_callback(
        responses.GET,
        endpoint,
        content_type="application/json",
        callback=list_callback,
    )


@pytest.fixture
def mock_thread(mocked_responses, api_url, account_id):
    base_thrd = {
        "id": "5678",
        "subject": "Test Thread",
        "account_id": account_id,
        "object": "thread",
        "folders": [{"name": "inbox", "display_name": "Inbox", "id": "abcd"}],
        "starred": True,
        "unread": False,
        "first_message_timestamp": 1451703845,
        "last_message_timestamp": 1483326245,
        "last_message_received_timestamp": 1483326245,
        "last_message_sent_timestamp": 1483232461,
    }
    response_body = json.dumps(base_thrd)

    def request_callback(request):
        payload = json.loads(request.body)
        if "folder" in payload:
            folder = {"name": "test", "display_name": "test", "id": payload["folder"]}
            base_thrd["folders"] = [folder]
        return (200, {}, json.dumps(base_thrd))

    endpoint = re.compile(api_url + "/threads/5678")
    mocked_responses.add(
        responses.GET,
        endpoint,
        content_type="application/json",
        status=200,
        body=response_body,
    )
    mocked_responses.add_callback(
        responses.PUT,
        endpoint,
        content_type="application/json",
        callback=request_callback,
    )


@pytest.fixture
def mock_labelled_thread(mocked_responses, api_url, account_id):
    base_thread = {
        "id": "111",
        "subject": "Labelled Thread",
        "account_id": account_id,
        "object": "thread",
        "folders": [{"name": "inbox", "display_name": "Inbox", "id": "abcd"}],
        "starred": True,
        "unread": False,
        "labels": [
            {
                "display_name": "Important",
                "id": "anuep8pe5ugmxrucchrzba2o8",
                "name": "important",
                "account_id": account_id,
                "object": "label",
            },
            {
                "display_name": "Existing",
                "id": "dfslhgy3rlijfhlsujnchefs3",
                "name": "existing",
                "account_id": account_id,
                "object": "label",
            },
        ],
        "first_message_timestamp": 1451703845,
        "last_message_timestamp": 1483326245,
        "last_message_received_timestamp": 1483326245,
        "last_message_sent_timestamp": 1483232461,
    }
    response_body = json.dumps(base_thread)

    def request_callback(request):
        payload = json.loads(request.body)
        if "labels" in payload:
            existing_labels = {label["id"]: label for label in base_thread["labels"]}
            new_labels = []
            for label_id in payload["labels"]:
                if label_id in existing_labels:
                    new_labels.append(existing_labels[label_id])
                else:
                    new_labels.append(
                        {
                            "name": "updated",
                            "display_name": "Updated",
                            "id": label_id,
                            "account_id": account_id,
                            "object": "label",
                        }
                    )
            copied = copy.copy(base_thread)
            copied["labels"] = new_labels
        return (200, {}, json.dumps(copied))

    endpoint = re.compile(api_url + "/threads/111")
    mocked_responses.add(
        responses.GET,
        endpoint,
        content_type="application/json",
        status=200,
        body=response_body,
    )
    mocked_responses.add_callback(
        responses.PUT,
        endpoint,
        content_type="application/json",
        callback=request_callback,
    )


@pytest.fixture
def mock_drafts(mocked_responses, api_url):
    drafts = [
        {
            "bcc": [],
            "body": "Cheers mate!",
            "cc": [],
            "date": 1438684486,
            "events": [],
            "files": [],
            "folder": None,
            "from": [],
            "id": "2h111aefv8pzwzfykrn7hercj",
            "namespace_id": "384uhp3aj8l7rpmv9s2y2rukn",
            "object": "draft",
            "reply_to": [],
            "reply_to_message_id": None,
            "snippet": "",
            "starred": False,
            "subject": "Here's an attachment",
            "thread_id": "clm33kapdxkposgltof845v9s",
            "to": [{"email": "helena@nylas.com", "name": "Helena Handbasket"}],
            "unread": False,
            "version": 0,
        }
    ]

    def list_callback(request):
        url = URLObject(request.url)
        offset = int(url.query_dict.get("offset") or 0)
        if offset:
            return (200, {}, json.dumps([]))
        return (200, {}, json.dumps(drafts))

    mocked_responses.add_callback(
        responses.GET,
        api_url + "/drafts",
        content_type="application/json",
        callback=list_callback,
    )


@pytest.fixture
def mock_draft_saved_response(mocked_responses, api_url):
    draft_json = {
        "bcc": [],
        "body": "Cheers mate!",
        "cc": [],
        "date": 1438684486,
        "events": [],
        "files": [],
        "folder": None,
        "from": [],
        "id": "2h111aefv8pzwzfykrn7hercj",
        "namespace_id": "384uhp3aj8l7rpmv9s2y2rukn",
        "object": "draft",
        "reply_to": [],
        "reply_to_message_id": None,
        "snippet": "",
        "starred": False,
        "subject": "Here's an attachment",
        "thread_id": "clm33kapdxkposgltof845v9s",
        "to": [{"email": "helena@nylas.com", "name": "Helena Handbasket"}],
        "unread": False,
        "version": 0,
    }

    def create_callback(_request):
        return (200, {}, json.dumps(draft_json))

    def update_callback(request):
        try:
            payload = json.loads(request.body)
        except ValueError:
            return (200, {}, json.dumps(draft_json))

        stripped_payload = {key: value for key, value in payload.items() if value}
        updated_draft_json = copy.copy(draft_json)
        updated_draft_json.update(stripped_payload)
        updated_draft_json["version"] += 1
        return (200, {}, json.dumps(updated_draft_json))

    mocked_responses.add_callback(
        responses.POST,
        api_url + "/drafts",
        content_type="application/json",
        callback=create_callback,
    )

    mocked_responses.add_callback(
        responses.PUT,
        api_url + "/drafts/2h111aefv8pzwzfykrn7hercj",
        content_type="application/json",
        callback=update_callback,
    )


@pytest.fixture
def mock_draft_deleted_response(mocked_responses, api_url):
    mocked_responses.add(
        responses.DELETE,
        api_url + "/drafts/2h111aefv8pzwzfykrn7hercj",
        content_type="application/json",
        status=200,
        body="",
    )


@pytest.fixture
def mock_draft_sent_response(mocked_responses, api_url):
    body = {
        "bcc": [],
        "body": "",
        "cc": [],
        "date": 1438684486,
        "events": [],
        "files": [],
        "folder": None,
        "from": [{"email": "benb@nylas.com"}],
        "id": "2h111aefv8pzwzfykrn7hercj",
        "namespace_id": "384uhp3aj8l7rpmv9s2y2rukn",
        "object": "draft",
        "reply_to": [],
        "reply_to_message_id": None,
        "snippet": "",
        "starred": False,
        "subject": "Stay polish, stay hungary",
        "thread_id": "clm33kapdxkposgltof845v9s",
        "to": [{"email": "helena@nylas.com", "name": "Helena Handbasket"}],
        "unread": False,
        "version": 0,
    }

    values = [(400, {}, "Couldn't send email"), (200, {}, json.dumps(body))]

    def callback(request):
        payload = json.loads(request.body)
        assert payload["draft_id"] == "2h111aefv8pzwzfykrn7hercj"
        return values.pop()

    mocked_responses.add_callback(
        responses.POST,
        api_url + "/send",
        callback=callback,
        content_type="application/json",
    )


@pytest.fixture
def mock_draft_send_unsaved_response(mocked_responses, api_url):
    def callback(request):
        payload = json.loads(request.body)
        payload["draft_id"] = "2h111aefv8pzwzfykrn7hercj"
        return 200, {}, json.dumps(payload)

    mocked_responses.add_callback(
        responses.POST,
        api_url + "/send",
        callback=callback,
        content_type="application/json",
    )


@pytest.fixture
def mock_files(mocked_responses, api_url, account_id):
    files_content = {"3qfe4k3siosfjtjpfdnon8zbn": b"Hello, World!"}
    files_metadata = {
        "3qfe4k3siosfjtjpfdnon8zbn": {
            "id": "3qfe4k3siosfjtjpfdnon8zbn",
            "content_type": "text/plain",
            "filename": "hello.txt",
            "account_id": account_id,
            "object": "file",
            "size": len(files_content["3qfe4k3siosfjtjpfdnon8zbn"]),
        }
    }
    mocked_responses.add(
        responses.GET,
        api_url + "/files",
        body=json.dumps(list(files_metadata.values())),
    )
    for file_id in files_content:
        mocked_responses.add(
            responses.POST,
            "{base}/files/{file_id}".format(base=api_url, file_id=file_id),
            body=json.dumps(files_metadata[file_id]),
        )
        mocked_responses.add(
            responses.GET,
            "{base}/files/{file_id}/download".format(base=api_url, file_id=file_id),
            body=files_content[file_id],
        )

    def create_callback(request):
        uploaded_lines = request.body.decode("utf8").splitlines()
        content_disposition = uploaded_lines[1]
        _, params = cgi.parse_header(content_disposition)
        filename = params.get("filename", None)
        content = "".join(uploaded_lines[3:-1])
        size = len(content.encode("utf8"))

        body = [
            {
                "id": generate_id(),
                "content_type": "text/plain",
                "filename": filename,
                "account_id": account_id,
                "object": "file",
                "size": size,
            }
        ]
        return (200, {}, json.dumps(body))

    mocked_responses.add_callback(
        responses.POST, api_url + "/files", callback=create_callback
    )


@pytest.fixture
def mock_event_create_response(mocked_responses, api_url, message_body):
    def callback(_request):
        try:
            payload = json.loads(_request.body)
        except ValueError:
            return 400, {}, ""

        payload["id"] = "cv4ei7syx10uvsxbs21ccsezf"
        return 200, {}, json.dumps(payload)

    mocked_responses.add_callback(
        responses.POST, api_url + "/events", callback=callback
    )

    put_body = {"title": "loaded from JSON", "ignored": "ignored"}
    mocked_responses.add(
        responses.PUT,
        api_url + "/events/cv4ei7syx10uvsxbs21ccsezf",
        body=json.dumps(put_body),
    )


@pytest.fixture
def mock_event_generate_ics(mocked_responses, api_url, message_body):
    mocked_responses.add(
        responses.POST, api_url + "/events/to-ics", body=json.dumps({"ics": ""})
    )


@pytest.fixture
def mock_scheduler_create_response(mocked_responses, api_url, message_body):
    def callback(_request):
        try:
            payload = json.loads(_request.body)
        except ValueError:
            return 400, {}, ""

        payload["id"] = "cv4ei7syx10uvsxbs21ccsezf"
        return 200, {}, json.dumps(payload)

    mocked_responses.add_callback(
        responses.POST, "https://api.schedule.nylas.com/manage/pages", callback=callback
    )

    mocked_responses.add(
        responses.PUT,
        "https://api.schedule.nylas.com/manage/pages/cv4ei7syx10uvsxbs21ccsezf",
        body=json.dumps(message_body),
    )


@pytest.fixture
def mock_event_create_response_with_limits(mocked_responses, api_url, message_body):
    def callback(request):
        url = URLObject(request.url)
        limit = int(url.query_dict.get("limit") or 50)
        body = [message_body for _ in range(0, limit)]
        return 200, {}, json.dumps(body)

    mocked_responses.add_callback(responses.GET, api_url + "/events", callback=callback)


@pytest.fixture
def mock_event_create_notify_response(mocked_responses, api_url, message_body):
    mocked_responses.add(
        responses.POST,
        api_url + "/events?notify_participants=true&other_param=1",
        body=json.dumps(message_body),
    )


@pytest.fixture
def mock_send_rsvp(mocked_responses, api_url, message_body):
    mocked_responses.add(
        responses.POST,
        re.compile(api_url + "/send-rsvp"),
        body=json.dumps(message_body),
    )


@pytest.fixture
def mock_components_create_response(mocked_responses, api_url, message_body):
    def callback(_request):
        try:
            payload = json.loads(_request.body)
        except ValueError:
            return 400, {}, ""

        payload["id"] = "cv4ei7syx10uvsxbs21ccsezf"
        return 200, {}, json.dumps(payload)

    mocked_responses.add_callback(
        responses.POST, re.compile(api_url + "/component/*"), callback=callback
    )

    mocked_responses.add(
        responses.PUT,
        re.compile(api_url + "/component/*"),
        body=json.dumps(message_body),
    )


@pytest.fixture
def mock_thread_search_response(mocked_responses, api_url):
    snippet = (
        "Hey Helena, Looking forward to getting together for dinner on Friday. "
        "What can I bring? I have a couple bottles of wine or could put together"
    )
    response_body = json.dumps(
        [
            {
                "id": "evh5uy0shhpm5d0le89goor17",
                "object": "thread",
                "account_id": "awa6ltos76vz5hvphkp8k17nt",
                "subject": "Dinner Party on Friday",
                "unread": False,
                "starred": False,
                "last_message_timestamp": 1398229259,
                "last_message_received_timestamp": 1398229259,
                "first_message_timestamp": 1298229259,
                "participants": [
                    {"name": "Ben Bitdiddle", "email": "ben.bitdiddle@gmail.com"}
                ],
                "snippet": snippet,
                "folders": [
                    {
                        "name": "inbox",
                        "display_name": "INBOX",
                        "id": "f0idlvozkrpj3ihxze7obpivh",
                    }
                ],
                "message_ids": [
                    "251r594smznew6yhiocht2v29",
                    "7upzl8ss738iz8xf48lm84q3e",
                    "ah5wuphj3t83j260jqucm9a28",
                ],
                "draft_ids": ["251r594smznew6yhi12312saq"],
                "version": 2,
            }
        ]
    )

    mocked_responses.add(
        responses.GET,
        api_url + "/threads/search?q=Helena",
        body=response_body,
        status=200,
        content_type="application/json",
        match_querystring=True,
    )


@pytest.fixture
def mock_message_search_response(mocked_responses, api_url):
    snippet = (
        "Sounds good--that bottle of Pinot should go well with the meal. "
        "I'll also bring a surprise for dessert. :) "
        "Do you have ice cream? Looking fo"
    )
    response_body = json.dumps(
        [
            {
                "id": "84umizq7c4jtrew491brpa6iu",
                "object": "message",
                "account_id": "14e5bn96uizyuhidhcw5rfrb0",
                "thread_id": "5vryyrki4fqt7am31uso27t3f",
                "subject": "Re: Dinner on Friday?",
                "from": [{"name": "Ben Bitdiddle", "email": "ben.bitdiddle@gmail.com"}],
                "to": [{"name": "Bill Rogers", "email": "wbrogers@mit.edu"}],
                "cc": [],
                "bcc": [],
                "reply_to": [],
                "date": 1370084645,
                "unread": True,
                "starred": False,
                "folder": {
                    "name": "inbox",
                    "display_name": "INBOX",
                    "id": "f0idlvozkrpj3ihxze7obpivh",
                },
                "snippet": snippet,
                "body": "<html><body>....</body></html>",
                "files": [],
                "events": [],
            },
            {
                "id": "84umizq7asdf3aw491brpa6iu",
                "object": "message",
                "account_id": "14e5bakdsfljskidhcw5rfrb0",
                "thread_id": "5vryyralskdjfwlj1uso27t3f",
                "subject": "Re: Dinner on Friday?",
                "from": [{"name": "Ben Bitdiddle", "email": "ben.bitdiddle@gmail.com"}],
                "to": [{"name": "Bill Rogers", "email": "wbrogers@mit.edu"}],
                "cc": [],
                "bcc": [],
                "reply_to": [],
                "date": 1370084645,
                "unread": True,
                "starred": False,
                "folder": {
                    "name": "inbox",
                    "display_name": "INBOX",
                    "id": "f0idlvozkrpj3ihxze7obpivh",
                },
                "snippet": snippet,
                "body": "<html><body>....</body></html>",
                "files": [],
                "events": [],
            },
        ]
    )

    mocked_responses.add(
        responses.GET,
        re.compile(api_url + "/messages/search\?q=Pinot.*"),
        body=response_body,
        status=200,
        content_type="application/json",
        match_querystring=True,
    )


@pytest.fixture
def mock_calendars(mocked_responses, api_url):
    calendars = [
        {
            "id": "8765",
            "events": [
                {
                    "title": "Pool party",
                    "location": "Local Community Pool",
                    "participants": ["Alice", "Bob", "Claire", "Dot"],
                }
            ],
        }
    ]

    def list_callback(request):
        url = URLObject(request.url)
        offset = int(url.query_dict.get("offset") or 0)
        if offset:
            return (200, {}, json.dumps([]))
        return (200, {}, json.dumps(calendars))

    endpoint = re.compile(api_url + "/calendars")
    mocked_responses.add_callback(
        responses.GET,
        endpoint,
        content_type="application/json",
        callback=list_callback,
    )


@pytest.fixture
def mock_contacts(mocked_responses, account_id, api_url):
    contact1 = {
        "id": "5x6b54whvcz1j22ggiyorhk9v",
        "object": "contact",
        "account_id": account_id,
        "given_name": "Charlie",
        "middle_name": None,
        "surname": "Bucket",
        "birthday": "1964-10-05",
        "suffix": None,
        "nickname": None,
        "company_name": None,
        "job_title": "Student",
        "manager_name": None,
        "office_location": None,
        "notes": None,
        "picture_url": "{base}/contacts/{id}/picture".format(
            base=api_url, id="5x6b54whvcz1j22ggiyorhk9v"
        ),
        "emails": [{"email": "charlie@gmail.com", "type": None}],
        "im_addresses": [],
        "physical_addresses": [],
        "phone_numbers": [],
        "web_pages": [],
    }
    contact2 = {
        "id": "4zqkfw8k1d12h0k784ipeh498",
        "object": "contact",
        "account_id": account_id,
        "given_name": "William",
        "middle_name": "J",
        "surname": "Wonka",
        "birthday": "1955-02-28",
        "suffix": None,
        "nickname": None,
        "company_name": None,
        "job_title": "Chocolate Artist",
        "manager_name": None,
        "office_location": "Willy Wonka Factory",
        "notes": None,
        "picture_url": None,
        "emails": [{"email": "scrumptious@wonka.com", "type": None}],
        "im_addresses": [],
        "physical_addresses": [],
        "phone_numbers": [],
        "web_pages": [{"type": "work", "url": "http://www.wonka.com"}],
    }
    contact3 = {
        "id": "9fn1aoi2i00qv6h1zpag6b26w",
        "object": "contact",
        "account_id": account_id,
        "given_name": "Oompa",
        "middle_name": None,
        "surname": "Loompa",
        "birthday": None,
        "suffix": None,
        "nickname": None,
        "company_name": None,
        "job_title": None,
        "manager_name": None,
        "office_location": "Willy Wonka Factory",
        "notes": None,
        "picture_url": None,
        "emails": [],
        "im_addresses": [],
        "physical_addresses": [],
        "phone_numbers": [],
        "web_pages": [],
    }
    contacts = [contact1, contact2, contact3]

    def list_callback(request):
        url = URLObject(request.url)
        offset = int(url.query_dict.get("offset") or 0)
        if offset:
            return (200, {}, json.dumps([]))
        return (200, {}, json.dumps(contacts))

    def create_callback(request):
        payload = json.loads(request.body)
        payload["id"] = generate_id()
        return (200, {}, json.dumps(payload))

    for contact in contacts:
        mocked_responses.add(
            responses.GET,
            re.compile(api_url + "/contacts/" + contact["id"]),
            content_type="application/json",
            status=200,
            body=json.dumps(contact),
        )
        if contact.get("picture_url"):
            mocked_responses.add(
                responses.GET,
                contact["picture_url"],
                content_type="image/jpeg",
                status=200,
                body=os.urandom(50),
                stream=True,
            )
        else:
            mocked_responses.add(
                responses.GET,
                "{base}/contacts/{id}/picture".format(base=api_url, id=contact["id"]),
                status=404,
                body="",
            )
    mocked_responses.add_callback(
        responses.GET,
        re.compile(api_url + "/contacts"),
        content_type="application/json",
        callback=list_callback,
    )
    mocked_responses.add_callback(
        responses.POST,
        api_url + "/contacts",
        content_type="application/json",
        callback=create_callback,
    )


@pytest.fixture
def mock_contact(mocked_responses, account_id, api_url):
    contact = {
        "id": "9hga75n6mdvq4zgcmhcn7hpys",
        "object": "contact",
        "account_id": account_id,
        "given_name": "Given",
        "middle_name": "Middle",
        "surname": "Sur",
        "birthday": "1964-10-05",
        "suffix": "Jr",
        "nickname": "Testy",
        "company_name": "Test Data Inc",
        "job_title": "QA Tester",
        "manager_name": "George",
        "office_location": "Over the Rainbow",
        "source": "inbox",
        "notes": "This is a note",
        "picture_url": "{base}/contacts/{id}/picture".format(
            base=api_url, id="9hga75n6mdvq4zgcmhcn7hpys"
        ),
        "emails": [
            {"type": "first", "email": "one@example.com"},
            {"type": "second", "email": "two@example.com"},
            {"type": "primary", "email": "abc@example.com"},
            {"type": "primary", "email": "xyz@example.com"},
            {"type": None, "email": "unknown@example.com"},
        ],
        "im_addresses": [
            {"type": "aim", "im_address": "SmarterChild"},
            {"type": "gtalk", "im_address": "fake@gmail.com"},
            {"type": "gtalk", "im_address": "fake2@gmail.com"},
        ],
        "physical_addresses": [
            {
                "type": "home",
                "format": "structured",
                "street_address": "123 Awesome Street",
                "postal_code": "99989",
                "state": "CA",
                "country": "America",
            }
        ],
        "phone_numbers": [
            {"type": "home", "number": "555-555-5555"},
            {"type": "mobile", "number": "555-555-5555"},
            {"type": "mobile", "number": "987654321"},
        ],
        "web_pages": [
            {"type": "profile", "url": "http://www.facebook.com/abc"},
            {"type": "profile", "url": "http://www.twitter.com/abc"},
            {"type": None, "url": "http://example.com"},
        ],
    }

    def update_callback(request):
        try:
            payload = json.loads(request.body)
        except ValueError:
            return (200, {}, json.dumps(contact))

        stripped_payload = {key: value for key, value in payload.items() if value}
        updated_contact_json = copy.copy(contact)
        updated_contact_json.update(stripped_payload)
        return (200, {}, json.dumps(updated_contact_json))

    mocked_responses.add(
        responses.GET,
        "{base}/contacts/{id}".format(base=api_url, id=contact["id"]),
        content_type="application/json",
        status=200,
        body=json.dumps(contact),
    )
    mocked_responses.add(
        responses.GET,
        contact["picture_url"],
        content_type="image/jpeg",
        status=200,
        body=os.urandom(50),
        stream=True,
    )

    mocked_responses.add_callback(
        responses.PUT,
        "{base}/contacts/{id}".format(base=api_url, id=contact["id"]),
        content_type="application/json",
        callback=update_callback,
    )


@pytest.fixture
def mock_events(mocked_responses, api_url):
    events = [
        {
            "id": "1234abcd5678",
            "message_id": "evh5uy0shhpm5d0le89goor17",
            "ical_uid": "19960401T080045Z-4000F192713-0052@example.com",
            "title": "Pool party",
            "location": "Local Community Pool",
            "participants": [
                {
                    "comment": None,
                    "email": "kelly@nylas.com",
                    "name": "Kelly Nylanaut",
                    "status": "noreply",
                },
                {
                    "comment": None,
                    "email": "sarah@nylas.com",
                    "name": "Sarah Nylanaut",
                    "status": "no",
                },
            ],
            "metadata": {},
        },
        {
            "id": "9876543cba",
            "message_id": None,
            "ical_uid": None,
            "title": "Event Without Message",
            "description": "This event does not have a corresponding message ID.",
            "metadata": {},
        },
        {
            "id": "1231241zxc",
            "message_id": None,
            "ical_uid": None,
            "title": "Event With Metadata",
            "description": "This event uses metadata to store custom values.",
            "metadata": {"platform": "python", "event_type": "meeting"},
        },
    ]

    def list_callback(request):
        url = URLObject(request.url)
        offset = int(url.query_dict.get("offset") or 0)
        metadata_key = url.query_multi_dict.get("metadata_key")
        metadata_value = url.query_multi_dict.get("metadata_value")
        metadata_pair = url.query_multi_dict.get("metadata_pair")

        if offset:
            return (200, {}, json.dumps([]))
        if metadata_key or metadata_value or metadata_pair:
            results = []
            for event in events:
                if (
                    metadata_key
                    and set(metadata_key) & set(event["metadata"])
                    or metadata_value
                    and set(metadata_value) & set(event["metadata"].values())
                ):
                    results.append(event)
                elif metadata_pair:
                    for pair in metadata_pair:
                        key_value = pair.split(":")
                        if (
                            key_value[0] in event["metadata"]
                            and event["metadata"][key_value[0]] == key_value[1]
                        ):
                            results.append(event)
            return (200, {}, json.dumps(results))
        return (200, {}, json.dumps(events))

    endpoint = re.compile(api_url + "/events")
    mocked_responses.add_callback(
        responses.GET, endpoint, content_type="application/json", callback=list_callback
    )


@pytest.fixture
def mock_schedulers(mocked_responses, api_url):
    scheduler_list = [
        {
            "app_client_id": "test-client-id",
            "app_organization_id": 12345,
            "config": {
                "appearance": {
                    "color": "#0068D3",
                    "company_name": "",
                    "logo": "",
                    "show_autoschedule": "true",
                    "show_nylas_branding": "false",
                    "show_timezone_options": "true",
                    "show_week_view": "true",
                    "submit_text": "Submit",
                },
                "locale": "en",
                "reminders": [],
                "timezone": "America/Los_Angeles",
            },
            "created_at": "2021-10-22",
            "edit_token": "test-edit-token-1",
            "id": 90210,
            "modified_at": "2021-10-22",
            "name": "test-1",
            "slug": "test1",
        },
        {
            "app_client_id": "test-client-id",
            "app_organization_id": 12345,
            "config": {
                "calendar_ids": {
                    "test-calendar-id": {
                        "availability": ["availability-id"],
                        "booking": "booking-id",
                    }
                },
                "event": {
                    "capacity": -1,
                    "duration": 45,
                    "location": "Location TBD",
                    "title": "test-event",
                },
                "locale": "en",
                "reminders": [],
                "timezone": "America/Los_Angeles",
            },
            "created_at": "2021-10-22",
            "edit_token": "test-edit-token-2",
            "id": 90211,
            "modified_at": "2021-10-22",
            "name": "test-2",
            "slug": "test2",
        },
    ]

    def list_callback(arg=None):
        return 200, {}, json.dumps(scheduler_list)

    def return_one_callback(arg=None):
        return 200, {}, json.dumps(scheduler_list[0])

    info_endpoint = re.compile("https://api.schedule.nylas.com/schedule/.*/info")

    mocked_responses.add_callback(
        responses.GET,
        "https://api.schedule.nylas.com/manage/pages",
        content_type="application/json",
        callback=list_callback,
    )

    mocked_responses.add_callback(
        responses.GET,
        info_endpoint,
        content_type="application/json",
        callback=return_one_callback,
    )


@pytest.fixture
def mock_scheduler_get_available_calendars(mocked_responses, api_url):
    calendars = [
        {
            "calendars": [
                {"id": "calendar-id", "name": "Emailed events", "read_only": "true"},
            ],
            "email": "swag@nylas.com",
            "id": "scheduler-id",
            "name": "Python Tester",
        }
    ]

    def list_callback(arg=None):
        return 200, {}, json.dumps(calendars)

    calendars_url = "https://api.schedule.nylas.com/manage/pages/{id}/calendars".format(
        id="cv4ei7syx10uvsxbs21ccsezf"
    )

    mocked_responses.add_callback(
        responses.GET,
        calendars_url,
        content_type="application/json",
        callback=list_callback,
    )


@pytest.fixture
def mock_scheduler_upload_image(mocked_responses, api_url):
    upload = {
        "filename": "test.png",
        "originalFilename": "test.png",
        "publicUrl": "https://public.nylas.com/test.png",
        "signedUrl": "https://signed.nylas.com/test.png",
    }

    def list_callback(arg=None):
        return 200, {}, json.dumps(upload)

    calendars_url = (
        "https://api.schedule.nylas.com/manage/pages/{id}/upload-image".format(
            id="cv4ei7syx10uvsxbs21ccsezf"
        )
    )

    mocked_responses.add_callback(
        responses.PUT,
        calendars_url,
        content_type="application/json",
        callback=list_callback,
    )


@pytest.fixture
def mock_scheduler_provider_availability(mocked_responses, api_url):
    response = {
        "busy": [
            {
                "end": 1636731958,
                "start": 1636728347,
            },
        ],
        "email": "test@example.com",
        "name": "John Doe",
    }

    def callback(arg=None):
        return 200, {}, json.dumps(response)

    provider_url = re.compile(
        "https://api.schedule.nylas.com/schedule/availability/(google|o365)"
    )

    mocked_responses.add_callback(
        responses.GET,
        provider_url,
        callback=callback,
    )


@pytest.fixture
def mock_scheduler_timeslots(mocked_responses, api_url):
    scheduler_time_slots = [
        {
            "account_id": "test-account-id",
            "calendar_id": "test-calendar-id",
            "emails": ["test@example.com"],
            "end": 1636731958,
            "host_name": "www.hostname.com",
            "start": 1636728347,
        },
    ]

    booking_confirmation = {
        "account_id": "test-account-id",
        "additional_field_values": {
            "test": "yes",
        },
        "calendar_event_id": "test-event-id",
        "calendar_id": "test-calendar-id",
        "edit_hash": "test-edit-hash",
        "end_time": 1636731958,
        "id": 123,
        "is_confirmed": False,
        "location": "Earth",
        "recipient_email": "recipient@example.com",
        "recipient_locale": "en_US",
        "recipient_name": "Recipient Doe",
        "recipient_tz": "America/New_York",
        "start_time": 1636728347,
        "title": "Test Booking",
    }

    cancel_payload = {
        "success": True,
    }

    def list_timeslots(arg=None):
        return 200, {}, json.dumps(scheduler_time_slots)

    def book_timeslot(arg=None):
        return 200, {}, json.dumps(booking_confirmation)

    def confirm_booking(arg=None):
        booking_confirmation["is_confirmed"] = True
        return 200, {}, json.dumps(booking_confirmation)

    def cancel_booking(arg=None):
        return 200, {}, json.dumps(cancel_payload)

    timeslots_url = re.compile("https://api.schedule.nylas.com/schedule/.*/timeslots")

    confirm_url = re.compile("https://api.schedule.nylas.com/schedule/.*/.*/confirm")

    cancel_url = re.compile("https://api.schedule.nylas.com/schedule/.*/.*/cancel")

    mocked_responses.add_callback(
        responses.GET,
        timeslots_url,
        callback=list_timeslots,
    )

    mocked_responses.add_callback(
        responses.POST,
        timeslots_url,
        callback=book_timeslot,
    )

    mocked_responses.add_callback(
        responses.POST,
        confirm_url,
        callback=confirm_booking,
    )

    mocked_responses.add_callback(
        responses.POST,
        cancel_url,
        callback=cancel_booking,
    )


@pytest.fixture
def mock_components(mocked_responses, api_url):
    components = [
        {
            "active": True,
            "settings": {},
            "allowed_domains": [],
            "id": "component-id",
            "name": "PyTest Component",
            "public_account_id": "account-id",
            "public_application_id": "application-id",
            "type": "agenda",
            "created_at": "2021-10-22T18:02:10.000Z",
            "updated_at": "2021-10-22T18:02:10.000Z",
            "accessed_at": None,
            "public_token_id": "token-id",
        },
    ]

    def list_callback(arg=None):
        return 200, {}, json.dumps(components)

    endpoint = re.compile(api_url + "/component/*")
    mocked_responses.add_callback(
        responses.GET, endpoint, content_type="application/json", callback=list_callback
    )


@pytest.fixture
def mock_create_webhook(mocked_responses, api_url, client_id):
    webhook = {"application_id": "application-id", "id": "webhook-id", "version": "1.0"}

    def callback(request):
        try:
            payload = json.loads(request.body)
        except ValueError:
            return 400, {}, ""

        if (
            "callback_url" not in payload
            and ("triggers" not in payload and type(payload["triggers"]) is not list)
            and "state" not in payload
        ):
            return 400, {}, ""

        webhook["callback_url"] = payload["callback_url"]
        webhook["triggers"] = payload["triggers"]
        webhook["state"] = payload["state"]

        return 200, {}, json.dumps(webhook)

    endpoint = "{base}/a/{client_id}/webhooks".format(base=api_url, client_id=client_id)
    mocked_responses.add_callback(
        responses.POST,
        endpoint,
        callback=callback,
        content_type="application/json",
    )


@pytest.fixture
def mock_webhooks(mocked_responses, api_url, client_id):
    webhook = {
        "application_id": "application-id",
        "callback_url": "https://your-server.com/webhook",
        "id": "webhook-id",
        "state": "active",
        "triggers": ["message.created"],
        "version": "2.0",
    }

    def list_callback(request):
        return 200, {}, json.dumps([webhook])

    def single_callback(request):
        webhook["id"] = get_id_from_url(request.url)
        return 200, {}, json.dumps(webhook)

    def update_callback(request):
        try:
            payload = json.loads(request.body)
        except ValueError:
            return 400, {}, ""

        if "state" in payload:
            webhook["state"] = payload["state"]
        webhook["id"] = get_id_from_url(request.url)
        return 200, {}, json.dumps(webhook)

    def delete_callback(request):
        return 200, {}, json.dumps({"success": True})

    def get_id_from_url(url):
        path = URLObject(url).path
        return path.rsplit("/", 1)[-1]

    endpoint_single = re.compile(
        "{base}/a/{client_id}/webhooks/*".format(base=api_url, client_id=client_id)
    )
    endpoint_list = "{base}/a/{client_id}/webhooks".format(
        base=api_url, client_id=client_id
    )
    mocked_responses.add_callback(
        responses.GET,
        endpoint_list,
        content_type="application/json",
        callback=list_callback,
    )
    mocked_responses.add_callback(
        responses.GET,
        endpoint_single,
        content_type="application/json",
        callback=single_callback,
    )
    mocked_responses.add_callback(
        responses.PUT,
        endpoint_single,
        content_type="application/json",
        callback=update_callback,
    )
    mocked_responses.add_callback(
        responses.DELETE,
        endpoint_single,
        content_type="application/json",
        callback=delete_callback,
    )


@pytest.fixture
def mock_resources(mocked_responses, api_url):
    resources = [
        {
            "object": "room_resource",
            "email": "training-room-1A@google.com",
            "name": "Google Training Room",
            "building": "San Francisco",
            "capacity": "10",
            "floor_name": "7",
            "floor_number": None,
        },
        {
            "object": "room_resource",
            "email": "training-room@outlook.com",
            "name": "Microsoft Training Room",
            "building": "Seattle",
            "capacity": "5",
            "floor_name": "Office",
            "floor_number": "2",
        },
    ]

    endpoint = re.compile(api_url + "/resources")
    mocked_responses.add(
        responses.GET,
        endpoint,
        body=json.dumps(resources),
        status=200,
        content_type="application/json",
    )


@pytest.fixture
def mock_job_statuses(mocked_responses, api_url):
    job_status = [
        {
            "account_id": "test_account_id",
            "action": "save_draft",
            "created_at": 1622846160,
            "id": "test_id",
            "job_status_id": "test_job_status_id",
            "object": "message",
            "status": "successful",
            "metadata": {"message_id": "nylas_message_id"},
        },
        {
            "account_id": "test_account_id",
            "action": "update_event",
            "created_at": 1622846160,
            "id": "test_id_2",
            "job_status_id": "test_job_status_id_2",
            "object": "event",
            "status": "successful",
        },
    ]

    endpoint = re.compile(api_url + "/job-statuses")
    mocked_responses.add(
        responses.GET,
        endpoint,
        body=json.dumps(job_status),
        status=200,
        content_type="application/json",
    )


@pytest.fixture
def mock_account_management(mocked_responses, api_url, account_id, client_id):
    account = {
        "account_id": account_id,
        "email_address": "ben.bitdiddle1861@gmail.com",
        "id": account_id,
        "name": "Ben Bitdiddle",
        "object": "account",
        "provider": "gmail",
        "organization_unit": "label",
        "billing_state": "paid",
        "authentication_type": "password",
    }
    paid_response = json.dumps(account)
    account["billing_state"] = "cancelled"
    cancelled_response = json.dumps(account)

    upgrade_url = "{base}/a/{client_id}/accounts/{id}/upgrade".format(
        base=api_url, id=account_id, client_id=client_id
    )
    downgrade_url = "{base}/a/{client_id}/accounts/{id}/downgrade".format(
        base=api_url, id=account_id, client_id=client_id
    )
    mocked_responses.add(
        responses.POST,
        upgrade_url,
        content_type="application/json",
        status=200,
        body=paid_response,
    )
    mocked_responses.add(
        responses.POST,
        downgrade_url,
        content_type="application/json",
        status=200,
        body=cancelled_response,
    )


@pytest.fixture
def mock_revoke_all_tokens(mocked_responses, api_url, account_id, client_id):
    revoke_all_url = "{base}/a/{client_id}/accounts/{id}/revoke-all".format(
        base=api_url, id=account_id, client_id=client_id
    )
    mocked_responses.add(
        responses.POST,
        revoke_all_url,
        content_type="application/json",
        status=200,
        body=json.dumps({"success": True}),
    )


@pytest.fixture
def mock_application_details(mocked_responses, api_url, client_id):
    application_details_url = "{base}/a/{client_id}".format(
        base=api_url, client_id=client_id
    )

    def modify_endpoint(request):
        return 200, {}, json.dumps(json.loads(request.body))

    mocked_responses.add(
        responses.GET,
        application_details_url,
        content_type="application/json",
        status=200,
        body=json.dumps(
            {
                "application_name": "My New App Name",
                "icon_url": "http://localhost:5555/icon.png",
                "redirect_uris": [
                    "http://localhost:5555/login_callback",
                    "localhost",
                    "https://customerA.myapplication.com/login_callback",
                ],
            }
        ),
    )
    mocked_responses.add_callback(
        responses.PUT,
        application_details_url,
        content_type="application/json",
        callback=modify_endpoint,
    )


@pytest.fixture
def mock_ip_addresses(mocked_responses, api_url, client_id):
    ip_addresses_url = "{base}/a/{client_id}/ip_addresses".format(
        base=api_url, client_id=client_id
    )
    mocked_responses.add(
        responses.GET,
        ip_addresses_url,
        content_type="application/json",
        status=200,
        body=json.dumps(
            {
                "ip_addresses": [
                    "39.45.235.23",
                    "23.10.341.123",
                    "12.56.256.654",
                    "67.20.987.231",
                ],
                "updated_at": 1552072984,
            }
        ),
    )


@pytest.fixture
def mock_token_info(mocked_responses, api_url, account_id, client_id):
    token_info_url = "{base}/a/{client_id}/accounts/{id}/token-info".format(
        base=api_url, id=account_id, client_id=client_id
    )
    mocked_responses.add(
        responses.POST,
        token_info_url,
        content_type="application/json",
        status=200,
        body=json.dumps(
            {
                "created_at": 1563496685,
                "scopes": "calendar,email,contacts",
                "state": "valid",
                "updated_at": 1563496685,
            }
        ),
    )


@pytest.fixture
def mock_free_busy(mocked_responses, api_url):
    free_busy_url = "{base}/calendars/free-busy".format(base=api_url)

    def free_busy_callback(request):
        payload = json.loads(request.body)
        email = payload["emails"][0]
        resp_data = [
            {
                "object": "free_busy",
                "email": email,
                "time_slots": [
                    {
                        "object": "time_slot",
                        "status": "busy",
                        "start_time": 1409594400,
                        "end_time": 1409598000,
                    },
                    {
                        "object": "time_slot",
                        "status": "busy",
                        "start_time": 1409598000,
                        "end_time": 1409599000,
                    },
                ],
            }
        ]
        return 200, {}, json.dumps(resp_data)

    mocked_responses.add_callback(
        responses.POST,
        free_busy_url,
        content_type="application/json",
        callback=free_busy_callback,
    )


@pytest.fixture
def mock_availability(mocked_responses, api_url):
    availability_url = "{base}/calendars/availability".format(base=api_url)

    def availability_callback(request):
        payload = json.loads(request.body)
        resp_data = {
            "object": "availability",
            "time_slots": [
                {
                    "object": "time_slot",
                    "status": "free",
                    "start_time": 1409594400,
                    "end_time": 1409598000,
                },
                {
                    "object": "time_slot",
                    "status": "free",
                    "start_time": 1409598000,
                    "end_time": 1409599000,
                },
            ],
        }

        return 200, {}, json.dumps(resp_data)

    mocked_responses.add_callback(
        responses.POST,
        availability_url,
        content_type="application/json",
        callback=availability_callback,
    )

    mocked_responses.add_callback(
        responses.POST,
        "{url}/consecutive".format(url=availability_url),
        content_type="application/json",
        callback=availability_callback,
    )


@pytest.fixture
def mock_sentiment_analysis(mocked_responses, api_url, account_id):
    sentiment_url = "{base}/neural/sentiment".format(base=api_url)

    def sentiment_callback(request):
        payload = json.loads(request.body)
        if "message_id" in payload:
            response = [
                {
                    "account_id": account_id,
                    "processed_length": 11,
                    "sentiment": "NEUTRAL",
                    "sentiment_score": 0.30000001192092896,
                    "text": "hello world",
                }
            ]
        else:
            response = {
                "account_id": account_id,
                "processed_length": len(payload["text"]),
                "sentiment": "NEUTRAL",
                "sentiment_score": 0.30000001192092896,
                "text": payload["text"],
            }

        return 200, {}, json.dumps(response)

    mocked_responses.add_callback(
        responses.PUT,
        sentiment_url,
        content_type="application/json",
        callback=sentiment_callback,
    )


@pytest.fixture
def mock_extract_signature(mocked_responses, api_url, account_id):
    signature_url = "{base}/neural/signature".format(base=api_url)

    def signature_callback(request):
        payload = json.loads(request.body)
        response = {
            "account_id": account_id,
            "body": "This is the body<div>Nylas Swag</div><div>Software Engineer</div><div>123-456-8901</div><div>swag@nylas.com</div><img src='https://example.com/logo.png' alt='https://example.com/link.html'></a>",
            "signature": "Nylas Swag\n\nSoftware Engineer\n\n123-456-8901\n\nswag@nylas.com",
            "date": 1624029503,
            "from": [
                {
                    "email": "swag@nylas.com",
                    "name": "Nylas Swag",
                },
            ],
            "id": "abc123",
            "model_version": "0.0.1",
            "object": "message",
            "provider_name": "gmail",
            "subject": "Subject",
            "to": [
                {
                    "email": "me@nylas.com",
                    "name": "me",
                },
            ],
        }
        if "parse_contacts" not in payload or payload["parse_contacts"] is True:
            response["contacts"] = {
                "job_titles": ["Software Engineer"],
                "links": [
                    {
                        "description": "string",
                        "url": "https://example.com/link.html",
                    },
                ],
                "phone_numbers": ["123-456-8901"],
                "emails": ["swag@nylas.com"],
                "names": [
                    {
                        "first_name": "Nylas",
                        "last_name": "Swag",
                    },
                ],
            }

        return 200, {}, json.dumps([response])

    mocked_responses.add_callback(
        responses.PUT,
        signature_url,
        content_type="application/json",
        callback=signature_callback,
    )


@pytest.fixture
def mock_categorize(mocked_responses, api_url, account_id):
    categorize_url = "{base}/neural/categorize".format(base=api_url)

    def categorize_callback(request):
        response = {
            "account_id": account_id,
            "body": "This is a body",
            "categorizer": {
                "categorized_at": 1627076720,
                "category": "feed",
                "model_version": "6194f733",
                "subcategories": ["ooo"],
            },
            "date": 1624029503,
            "from": [
                {
                    "email": "swag@nylas.com",
                    "name": "Nylas Swag",
                },
            ],
            "id": "abc123",
            "object": "message",
            "provider_name": "gmail",
            "subject": "Subject",
            "to": [
                {
                    "email": "me@nylas.com",
                    "name": "me",
                },
            ],
        }

        return 200, {}, json.dumps([response])

    def recategorize_callback(request):
        response = {
            "account_id": account_id,
            "category": "conversation",
            "is_primary_label": "true",
            "message_id": "abc123",
            "recategorized_at": "2021-07-17T00:04:22.006193",
            "recategorized_from": {
                "category": "feed",
                "model_version": "6194f733",
                "subcategories": ["ooo"],
            },
            "subcategories": ["ooo"],
        }

        return 200, {}, json.dumps(response)

    mocked_responses.add_callback(
        responses.PUT,
        categorize_url,
        content_type="application/json",
        callback=categorize_callback,
    )

    mocked_responses.add_callback(
        responses.POST,
        "{}/feedback".format(categorize_url),
        content_type="application/json",
        callback=recategorize_callback,
    )


@pytest.fixture
def mock_ocr_request(mocked_responses, api_url, account_id):
    ocr_url = "{base}/neural/ocr".format(base=api_url)

    def ocr_callback(request):
        response = {
            "account_id": account_id,
            "content_type": "application/pdf",
            "filename": "sample.pdf",
            "id": "abc123",
            "object": "file",
            "ocr": ["This is page 1", "This is page 2"],
            "processed_pages": 2,
            "size": 20,
        }

        return 200, {}, json.dumps(response)

    mocked_responses.add_callback(
        responses.PUT,
        ocr_url,
        content_type="application/json",
        callback=ocr_callback,
    )


@pytest.fixture
def mock_clean_conversation(mocked_responses, api_url, account_id):
    conversation_url = "{base}/neural/conversation".format(base=api_url)
    file_url = "{base}/files/1781777f666586677621".format(base=api_url)

    def conversation_callback(request):
        response = {
            "account_id": account_id,
            "body": "<img src='cid:1781777f666586677621' /> This is the body",
            "conversation": "<img src='cid:1781777f666586677621' /> This is the conversation",
            "date": 1624029503,
            "from": [
                {
                    "email": "swag@nylas.com",
                    "name": "Nylas Swag",
                },
            ],
            "id": "abc123",
            "model_version": "0.0.1",
            "object": "message",
            "provider_name": "gmail",
            "subject": "Subject",
            "to": [
                {
                    "email": "me@nylas.com",
                    "name": "me",
                },
            ],
        }

        return 200, {}, json.dumps([response])

    def file_callback(request):
        response = {
            "id": "1781777f666586677621",
            "content_type": "image/png",
            "filename": "hello.png",
            "account_id": account_id,
            "object": "file",
            "size": 123,
        }

        return 200, {}, json.dumps(response)

    mocked_responses.add_callback(
        responses.PUT,
        conversation_url,
        content_type="application/json",
        callback=conversation_callback,
    )

    mocked_responses.add_callback(
        responses.GET,
        file_url,
        content_type="application/json",
        callback=file_callback,
    )


@pytest.fixture
def mock_deltas_since(mocked_responses, api_url):
    deltas = {
        "cursor_start": "start_cursor",
        "cursor_end": "end_cursor",
        "deltas": [
            {
                "attributes": {
                    "account_id": "aid-5678",
                    "given_name": "First",
                    "surname": "Last",
                    "id": "id-1234",
                    "object": "contact",
                },
                "cursor": "contact_cursor",
                "event": "create",
                "id": "delta-1",
                "object": "contact",
            },
            {
                "attributes": {
                    "account_id": "aid-5678",
                    "content_type": "text/plain",
                    "filename": "sample.txt",
                    "id": "id-1234",
                    "object": "file",
                    "size": 123,
                },
                "cursor": "file_cursor",
                "event": "create",
                "id": "delta-2",
                "object": "file",
            },
            {
                "attributes": {
                    "account_id": "aid-5678",
                    "to": [{"email": "foo", "name": "bar"}],
                    "subject": "foo",
                    "id": "id-1234",
                    "object": "message",
                },
                "cursor": "message_cursor",
                "event": "create",
                "id": "delta-3",
                "object": "message",
            },
            {
                "attributes": {
                    "account_id": "aid-5678",
                    "to": [{"email": "foo", "name": "bar"}],
                    "subject": "foo",
                    "id": "id-1234",
                    "object": "draft",
                },
                "cursor": "draft_cursor",
                "event": "create",
                "id": "delta-4",
                "object": "draft",
            },
            {
                "attributes": {
                    "account_id": "aid-5678",
                    "subject": "Subject",
                    "id": "id-1234",
                    "object": "thread",
                },
                "cursor": "thread_cursor",
                "event": "create",
                "id": "delta-5",
                "object": "thread",
            },
            {
                "attributes": {
                    "id": "id-1234",
                    "title": "test event",
                    "when": {"time": 1409594400, "object": "time"},
                    "participants": [
                        {
                            "name": "foo",
                            "email": "bar",
                            "status": "noreply",
                            "comment": "This is a comment",
                            "phone_number": "416-000-0000",
                        },
                    ],
                    "ical_uid": "id-5678",
                    "master_event_id": "master-1234",
                    "original_start_time": 1409592400,
                },
                "cursor": "event_cursor",
                "event": "create",
                "id": "delta-6",
                "object": "event",
            },
            {
                "attributes": {
                    "account_id": "aid-5678",
                    "id": "id-1234",
                    "object": "folder",
                    "name": "inbox",
                    "display_name": "name",
                },
                "cursor": "folder_cursor",
                "event": "create",
                "id": "delta-7",
                "object": "folder",
            },
            {
                "attributes": {
                    "account_id": "aid-5678",
                    "id": "id-1234",
                    "object": "label",
                    "name": "inbox",
                },
                "cursor": "label_cursor",
                "event": "create",
                "id": "delta-8",
                "object": "label",
            },
        ],
    }

    def callback(request):
        return 200, {}, json.dumps(deltas)

    mocked_responses.add_callback(
        responses.GET,
        "{base}/delta".format(base=api_url),
        callback=callback,
        content_type="application/json",
    )


@pytest.fixture
def mock_delta_cursor(mocked_responses, api_url):
    def callback(request):
        return 200, {}, json.dumps({"cursor": "cursor"})

    mocked_responses.add_callback(
        responses.POST,
        "{base}/delta/latest_cursor".format(base=api_url),
        callback=callback,
        content_type="application/json",
    )


@pytest.fixture
def mock_delta_stream(mocked_responses, api_url):
    delta = {
        "attributes": {
            "account_id": "aid-5678",
            "given_name": "First",
            "surname": "Last",
            "id": "id-1234",
            "object": "contact",
        },
        "cursor": "contact_cursor",
        "event": "create",
        "id": "delta-1",
        "object": "contact",
    }

    def stream_callback(request):
        return 200, {}, json.dumps(delta)

    def longpoll_callback(request):
        response = {
            "cursor_start": "start_cursor",
            "cursor_end": "end_cursor",
            "deltas": [delta],
        }
        return 200, {}, json.dumps(response)

    mocked_responses.add_callback(
        responses.GET,
        "{base}/delta/streaming".format(base=api_url),
        callback=stream_callback,
        content_type="application/json",
    )

    mocked_responses.add_callback(
        responses.GET,
        "{base}/delta/longpoll".format(base=api_url),
        callback=longpoll_callback,
        content_type="application/json",
    )


@pytest.fixture
def mock_outbox(mocked_responses, api_url):
    outbox_job_status = {
        "job_status_id": "job-status-id",
        "status": "pending",
        "original_data": {
            "subject": "With Love, from Nylas",
            "to": [{"name": "Me", "email": "test@email.com"}],
            "body": "This email was sent using the Nylas email API. Visit https://nylas.com for details.",
        },
        "account_id": "account-id",
    }

    def return_job_status(request):
        response = outbox_job_status
        payload = json.loads(request.body)
        if "send_at" in payload:
            response["original_data"]["send_at"] = payload["send_at"]
            response["original_data"]["original_send_at"] = payload["send_at"]
            response["original_data"]["retry_limit_datetime"] = payload["send_at"]
        if "retry_limit_datetime" in payload:
            response["original_data"]["retry_limit_datetime"] = payload[
                "retry_limit_datetime"
            ]
        return 200, {}, json.dumps(response)

    def delete_callback(request):
        return 200, {}, ""

    outbox_endpoint = "{base}/v2/outbox".format(base=api_url)
    endpoint_single = re.compile("{outbox_url}/*".format(outbox_url=outbox_endpoint))

    mocked_responses.add_callback(
        responses.POST,
        outbox_endpoint,
        callback=return_job_status,
        content_type="application/json",
    )

    mocked_responses.add_callback(
        responses.PATCH,
        endpoint_single,
        callback=return_job_status,
        content_type="application/json",
    )

    mocked_responses.add_callback(
        responses.DELETE,
        endpoint_single,
        callback=delete_callback,
        content_type="application/json",
    )


@pytest.fixture
def mock_outbox_send_grid(mocked_responses, api_url):
    send_grid_verification = {
        "results": {"domain_verified": True, "sender_verified": True}
    }

    def return_status(request):
        return 200, {}, json.dumps(send_grid_verification)

    def delete_callback(request):
        return 200, {}, ""

    verification_url = "{base}/v2/outbox/onboard/verified_status".format(base=api_url)
    delete_url = "{base}/v2/outbox/onboard/subuser".format(base=api_url)

    mocked_responses.add_callback(
        responses.GET,
        verification_url,
        callback=return_status,
        content_type="application/json",
    )

    mocked_responses.add_callback(
        responses.DELETE,
        delete_url,
        callback=delete_callback,
        content_type="application/json",
    )


@pytest.fixture
def mock_integrations(mocked_responses, client_id):
    integration = {
        "name": "Nylas Playground",
        "provider": "zoom",
        "settings": {
            "client_id": "test_client_id",
            "client_secret": "test_client_secret",
        },
        "redirect_uris": ["https://www.nylas.com"],
        "expires_in": 12000,
    }

    def list_callback(request):
        response = {"data": [integration], "limit": 10, "offset": 0}
        return 200, {}, json.dumps(response)

    def single_callback(request):
        integration["provider"] = get_id_from_url(request.url)
        response = {"data": integration}
        return 200, {}, json.dumps(response)

    def update_callback(request):
        try:
            payload = json.loads(request.body)
        except ValueError:
            return 400, {}, ""

        response = {"success": True, "data": payload}
        return 200, {}, json.dumps(response)

    def delete_callback(request):
        return 200, {}, json.dumps({"success": True})

    def get_id_from_url(url):
        path = URLObject(url).path
        return path.rsplit("/", 1)[-1]

    endpoint_post = re.compile("https://.*nylas.com/connect/integrations")
    endpoint_single = re.compile("https://.*nylas.com/connect/integrations/.*")
    endpoint_list = re.compile("https://.*nylas.com/connect/integrations\?.*")
    mocked_responses.add_callback(
        responses.GET,
        endpoint_list,
        content_type="application/json",
        callback=list_callback,
    )
    mocked_responses.add_callback(
        responses.GET,
        endpoint_single,
        content_type="application/json",
        callback=single_callback,
    )
    mocked_responses.add_callback(
        responses.POST,
        endpoint_post,
        content_type="application/json",
        callback=update_callback,
    )
    mocked_responses.add_callback(
        responses.PATCH,
        endpoint_single,
        content_type="application/json",
        callback=update_callback,
    )
    mocked_responses.add_callback(
        responses.DELETE,
        endpoint_single,
        content_type="application/json",
        callback=delete_callback,
    )


@pytest.fixture
def mock_grants(mocked_responses, client_id):
    grant = {
        "id": "grant-id",
        "provider": "zoom",
        "grant_status": "valid",
        "email": "email@example.com",
        "metadata": {"isAdmin": True},
        "scope": ["meeting:write"],
        "user_agent": "string",
        "ip": "string",
        "state": "my-state",
        "created_at": 1617817109,
        "updated_at": 1617817109,
    }

    def list_callback(request):
        response = {"data": [grant], "limit": 10, "offset": 0}
        return 200, {}, json.dumps(response)

    def single_callback(request):
        response = {"data": grant}
        return 200, {}, json.dumps(response)

    def update_callback(request):
        try:
            payload = json.loads(request.body)
        except ValueError:
            return 400, {}, ""

        response = {"success": True, "data": payload}
        return 200, {}, json.dumps(response)

    def delete_callback(request):
        return 200, {}, json.dumps({"success": True})

    def on_demand_sync(request):
        return 200, {}, json.dumps(grant)

    endpoint_post = re.compile("https://.*nylas.com/connect/grants")
    endpoint_single = re.compile("https://.*nylas.com/connect/grants/.*")
    endpoint_list = re.compile("https://.*nylas.com/connect/grants\?.*")
    endpoint_sync = re.compile("https://.*nylas.com/connect/grants/.*/sync.*")
    mocked_responses.add_callback(
        responses.POST,
        endpoint_sync,
        content_type="application/json",
        callback=on_demand_sync,
    )
    mocked_responses.add_callback(
        responses.GET,
        endpoint_list,
        content_type="application/json",
        callback=list_callback,
    )
    mocked_responses.add_callback(
        responses.GET,
        endpoint_single,
        content_type="application/json",
        callback=single_callback,
    )
    mocked_responses.add_callback(
        responses.POST,
        endpoint_post,
        content_type="application/json",
        callback=update_callback,
    )
    mocked_responses.add_callback(
        responses.PATCH,
        endpoint_single,
        content_type="application/json",
        callback=update_callback,
    )
    mocked_responses.add_callback(
        responses.DELETE,
        endpoint_single,
        content_type="application/json",
        callback=delete_callback,
    )


@pytest.fixture
def mock_authentication_hosted_auth(mocked_responses, client_id):
    api_response = {
        "success": True,
        "data": {
            "url": "https://accounts.nylas.com/connect/login?id=uas-hosted-id",
            "id": "uas-hosted-id",
            "expires_at": 0,
            "request": {},
        },
    }

    def hosted_auth_response(request):
        try:
            payload = json.loads(request.body)
        except ValueError:
            return 400, {}, ""

        api_response["data"]["request"] = payload
        return 200, {}, json.dumps(api_response)

    endpoint = re.compile("https://.*nylas.com/connect/auth")
    mocked_responses.add_callback(
        responses.POST,
        endpoint,
        content_type="application/json",
        callback=hosted_auth_response,
    )
